# Copyright 2015 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
# Copyright 2016 - Brocade Communications Systems, Inc.
# Copyright 2020 Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import contextlib
import datetime
import re
import sys
import threading

from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_db import sqlalchemy as oslo_sqlalchemy
from oslo_db.sqlalchemy import utils as db_utils
from oslo_log import log as logging
from oslo_utils import uuidutils  # noqa
import sqlalchemy as sa

from mistral import context
from mistral.db.sqlalchemy import base as b
from mistral.db.sqlalchemy import model_base as mb
from mistral.db.sqlalchemy import sqlite_lock
from mistral.db import utils as m_dbutils
from mistral.db.v2.sqlalchemy import filters as db_filters
from mistral.db.v2.sqlalchemy import models
from mistral import exceptions as exc
from mistral.services import security
from mistral.workflow import states
from mistral_lib import utils

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

_SCHEMA_LOCK = threading.RLock()
_initialized = False


def get_backend():
    """Consumed by openstack common code.

    The backend is this module itself.
    :return: Name of db backend.
    """
    return sys.modules[__name__]


def setup_db():
    global _initialized

    with _SCHEMA_LOCK:
        if _initialized:
            return

        try:
            models.Workbook.metadata.create_all(b.get_engine())

            _initialized = True
        except sa.exc.OperationalError as e:
            raise exc.DBError("Failed to setup database: %s" % e)


def drop_db():
    global _initialized

    with _SCHEMA_LOCK:
        if not _initialized:
            return

        try:
            models.Workbook.metadata.drop_all(b.get_engine())

            _initialized = False
        except Exception as e:
            raise exc.DBError("Failed to drop database: %s" % e)


# Transaction management.

def start_tx():
    b.start_tx()


def commit_tx():
    b.commit_tx()


def rollback_tx():
    b.rollback_tx()


def end_tx():
    b.end_tx()


@contextlib.contextmanager
def transaction(read_only=False):
    start_tx()

    try:
        yield
        if read_only:
            rollback_tx()
        else:
            commit_tx()
    finally:
        end_tx()


@b.session_aware()
def refresh(model, session=None):
    session.refresh(model)


@b.session_aware()
def expire_all(session=None):
    session.expire_all()


@b.session_aware()
def acquire_lock(model, id, session=None):
    # Expire all so all objects queried after lock is acquired
    # will be up-to-date from the DB and not from cache.
    session.expire_all()

    if b.get_driver_name() == 'sqlite':
        # In case of 'sqlite' we need to apply a manual lock.
        sqlite_lock.acquire_lock(id, session)

    return _lock_entity(model, id)


def _lock_entity(model, id):
    # Get entity by ID in "FOR UPDATE" mode and expect exactly one object.
    return _secure_query(model).with_for_update().filter(model.id == id).one()


@b.session_aware()
def update_on_match(id, specimen, values, attempts, session=None):
    """Updates a model with the given values if it matches the given specimen.

    :param id: ID of a persistent model.
    :param specimen: Specimen used to match the
    :param values: Values to set to the model if fields of the object
        match the specimen.
    :param attempts: The function will then invoke the UPDATE statement and
        check for "success" one or more times, up to a maximum of that passed
        as attempts.
    :param session: Session.
    :return: Persistent object attached to the session.
    """

    assert id is not None
    assert specimen is not None

    # We need to flush the session because when we do update_on_match()
    # it doesn't always update the state of the persistent object properly
    # when it merges a specimen state into it. Some fields get wiped out from
    # the history of ORM events that must be flushed later. For example, it
    # doesn't work well in case of Postgres.
    # See https://bugs.launchpad.net/mistral/+bug/1736821
    session.flush()

    model = None
    model_class = type(specimen)

    # Use WHERE clause to exclude possible conflicts if the state has
    # already been changed.
    try:
        model = b.model_query(model_class).update_on_match(
            specimen=specimen,
            surrogate_key='id',
            values=values,
            attempts=attempts
        )
    except oslo_sqlalchemy.update_match.NoRowsMatched:
        LOG.info(
            "Can't change state of persistent object "
            "because it has already been changed. [model_class=%s, id=%s, "
            "specimen=%s, values=%s]",
            model_class, id, specimen, values
        )

    return model


def _secure_query(model, *columns):
    query = b.model_query(model, columns)

    if not issubclass(model, mb.MistralSecureModelBase):
        return query

    shared_res_ids = []
    res_type = RESOURCE_MAPPING.get(model, '')

    if res_type:
        shared_res = _get_accepted_resources(res_type)
        shared_res_ids = [res.resource_id for res in shared_res]

    query_criterion = sa.or_(
        model.project_id == security.get_project_id(),
        model.scope == 'public'
    )

    # NOTE(kong): Include IN_ predicate in query filter only if shared_res_ids
    # is not empty to avoid sqlalchemy SAWarning and wasting a db call.
    if shared_res_ids:
        query_criterion = sa.or_(
            query_criterion,
            model.id.in_(shared_res_ids)
        )

    query = query.filter(query_criterion)

    return query


def _paginate_query(model, limit=None, marker=None, sort_keys=None,
                    sort_dirs=None, query=None):
    if not query:
        query = _secure_query(model)

    sort_keys = sort_keys if sort_keys else []

    # We should add sorting by id only if we use pagination or when
    # there is no specified ordering criteria. Otherwise
    # we can omit it to increase the performance.
    if not sort_keys or (marker or limit) and 'id' not in sort_keys:
        sort_keys.append('id')
        sort_dirs.append('asc') if sort_dirs else None

    query = db_utils.paginate_query(
        query,
        model,
        limit,
        sort_keys,
        marker=marker,
        sort_dirs=sort_dirs
    )

    return query


def _delete_all(model, **kwargs):
    # NOTE(kong): Because we use 'in_' operator in _secure_query(), delete()
    # method will raise error with default parameter. Please refer to
    # http://docs.sqlalchemy.org/en/rel_1_0/orm/query.html#sqlalchemy.orm.query.Query.delete
    query = _secure_query(model)
    query = db_filters.apply_filters(query, model, **kwargs)
    query.delete(synchronize_session=False)


def _get_collection(model, insecure=False, limit=None, marker=None,
                    sort_keys=None, sort_dirs=None, fields=None, **filters):
    columns = (
        tuple([getattr(model, f) for f in fields if hasattr(model, f)])
        if fields else ()
    )

    query = (b.model_query(model, columns=columns) if insecure
             else _secure_query(model, *columns))
    query = db_filters.apply_filters(query, model, **filters)

    query = _paginate_query(
        model,
        limit,
        marker,
        sort_keys,
        sort_dirs,
        query
    )

    return query.all()


def get_db_objects(model, insecure=False, **filters):
    query = b.model_query(model) if insecure else _secure_query(model)
    query = db_filters.apply_filters(query, model, **filters)

    return query


def _get_count(model, insecure=False, **filters):
    query = b.model_query(model) if insecure else _secure_query(model)

    query = db_filters.apply_filters(query, model, **filters)

    return query.count()


def _get_db_object_by_name(model, name, columns=()):
    query = _secure_query(model, *columns)

    return query.filter_by(name=name).first()


def _get_db_object_by_id(model, id, insecure=False, columns=()):
    columns = (
        tuple([getattr(model, f) for f in columns if hasattr(model, f)])
        if columns and isinstance(columns, list) else columns
    )

    query = (
        b.model_query(model, columns=columns)
        if insecure
        else _secure_query(model, *columns)
    )

    return query.filter_by(id=id).first()


def _get_db_object_by_name_and_namespace_or_id(model, identifier,
                                               namespace=None, insecure=False,
                                               columns=()):
    columns = (
        tuple([getattr(model, f) for f in columns if hasattr(model, f)])
        if columns and isinstance(columns, list) else columns
    )
    query = (
        b.model_query(model, columns=columns)
        if insecure
        else _secure_query(model, *columns)
    )

    match_name = model.name == identifier

    if namespace is not None:
        match_name = sa.and_(match_name, model.namespace == namespace)

    match_id = model.id == identifier

    query = query.filter(
        sa.or_(
            match_id,
            match_name
        )
    )

    return query.first()


def _get_db_object_by_name_and_namespace(model, name,
                                         namespace='', insecure=False,
                                         columns=()):
    query = (
        b.model_query(model, columns=columns)
        if insecure
        else _secure_query(model, *columns)
    )

    if namespace is None:
        namespace = ''

    query = query.filter(
        sa.and_(
            model.name == name,
            model.namespace == namespace
        )
    )

    return query.first()


# Workbook definitions.

@b.session_aware()
def get_workbook(name, namespace=None, fields=(), session=None):
    wb = _get_db_object_by_name_and_namespace(
        models.Workbook,
        name,
        namespace,
        columns=fields
    )

    if not wb:
        raise exc.DBEntityNotFoundError(
            "Workbook not found [name=%s, namespace=%s]" % (name, namespace)
        )

    return wb


@b.session_aware()
def load_workbook(name, namespace=None, fields=(), session=None):
    return _get_db_object_by_name_and_namespace(
        models.Workbook,
        name,
        namespace,
        columns=fields
    )


@b.session_aware()
def get_workbooks(session=None, **kwargs):
    return _get_collection(models.Workbook, **kwargs)


@b.session_aware()
def create_workbook(values, session=None):
    wb = models.Workbook()

    wb.update(values.copy())

    try:
        wb.save(session=session)
    except db_exc.DBDuplicateEntry:
        raise exc.DBDuplicateEntryError(
            "Duplicate entry for WorkbookDefinition "
            "['name', 'namespace', 'project_id']: {}, {}, {}".format(
                wb.name, wb.namespace, wb.project_id)
        )

    return wb


@b.session_aware()
def update_workbook(name, values, session=None):
    namespace = values.get('namespace')
    wb = get_workbook(name, namespace=namespace)

    wb.update(values.copy())

    return wb


@b.session_aware()
def create_or_update_workbook(name, values, session=None):
    if not _get_db_object_by_name(models.Workbook, name):
        return create_workbook(values)
    else:
        return update_workbook(name, values)


@b.session_aware()
def delete_workbook(name, namespace=None, session=None):
    namespace = namespace or ''

    count = _secure_query(models.Workbook).filter(
        sa.and_(
            models.Workbook.name == name,
            models.Workbook.namespace == namespace
        )
    ).delete()

    if count == 0:
        raise exc.DBEntityNotFoundError(
            "Workbook not found [workbook_name=%s, namespace=%s]"
            % (name, namespace)
        )


@b.session_aware()
def delete_workbooks(session=None, **kwargs):
    return _delete_all(models.Workbook, **kwargs)


# Workflow definitions.

@b.session_aware()
def get_workflow_definition(identifier, namespace='', fields=(), session=None):
    """Gets workflow definition by name or uuid.

    :param identifier: Identifier could be in the format of plain string or
                       uuid.
    :param namespace: The namespace the workflow is in. Optional.
    :param fields: Fields that need to be loaded. For example,
        (WorkflowDefinition.name,)
    :return: Workflow definition.
    """
    ctx = context.ctx()

    wf_def = _get_db_object_by_name_and_namespace_or_id(
        models.WorkflowDefinition,
        identifier,
        namespace=namespace,
        insecure=ctx.is_admin,
        columns=fields
    )

    if not wf_def:
        raise exc.DBEntityNotFoundError(
            "Workflow not found [workflow_identifier=%s, namespace=%s]"
            % (identifier, namespace)
        )

    return wf_def


@b.session_aware()
def get_workflow_definition_by_id(id, fields=(), session=None):
    wf_def = _get_db_object_by_id(
        models.WorkflowDefinition,
        id,
        columns=fields
    )

    if not wf_def:
        raise exc.DBEntityNotFoundError(
            "Workflow not found [workflow_id=%s]" % id
        )

    return wf_def


@b.session_aware()
def load_workflow_definition(name, namespace='', fields=(), session=None):
    model = models.WorkflowDefinition

    query = _secure_query(model, *fields)

    filter_ = sa.and_(
        model.name == name, model.namespace.in_([namespace, ''])
    )

    # Give priority to objects not in the default namespace.
    order_by = model.namespace.desc()

    if order_by is not None:
        query = query.order_by(order_by)

    return query.filter(filter_).first()


@b.session_aware()
def get_workflow_definitions(fields=None, session=None, **kwargs):
    if fields and 'input' in fields:
        fields.remove('input')
        fields.append('spec')

    return _get_collection(
        model=models.WorkflowDefinition,
        fields=fields,
        **kwargs
    )


@b.session_aware()
def create_workflow_definition(values, session=None):
    wf_def = models.WorkflowDefinition()

    wf_def.update(values.copy())

    try:
        wf_def.save(session=session)
    except db_exc.DBDuplicateEntry:
        raise exc.DBDuplicateEntryError(
            "Duplicate entry for WorkflowDefinition ['name', 'namespace',"
            " 'project_id']: {}, {}, {}".format(wf_def.name, wf_def.namespace,
                                                wf_def.project_id))
    return wf_def


@b.session_aware()
def update_workflow_definition(identifier, values, session=None):
    namespace = values.get('namespace')
    wf_def = get_workflow_definition(identifier, namespace=namespace)

    m_dbutils.check_db_obj_access(wf_def)

    if wf_def.scope == 'public' and values['scope'] == 'private':
        # Check cron triggers.
        cron_triggers = get_cron_triggers(insecure=True, workflow_id=wf_def.id)

        for c_t in cron_triggers:
            if c_t.project_id != wf_def.project_id:
                raise exc.NotAllowedException(
                    "Can not update scope of workflow that has cron triggers "
                    "associated in other tenants. [workflow_identifier=%s, "
                    "namespace=%s]" % (identifier, namespace)
                )

        # Check event triggers.
        event_triggers = get_event_triggers(
            insecure=True,
            workflow_id=wf_def.id
        )

        for e_t in event_triggers:
            if e_t.project_id != wf_def.project_id:
                raise exc.NotAllowedException(
                    "Can not update scope of workflow that has event triggers "
                    "associated in other tenants. [workflow_identifier=%s, "
                    "namespace=%s]" % (identifier, namespace)
                )

    wf_def.update(values.copy())

    return wf_def


@b.session_aware()
def create_or_update_workflow_definition(name, values, session=None):
    namespace = values.get('namespace')
    if _get_db_object_by_name_and_namespace_or_id(
            models.WorkflowDefinition,
            name,
            namespace=namespace):
        return update_workflow_definition(name, values)
    return create_workflow_definition(values)


@b.session_aware()
def delete_workflow_definition(identifier, namespace='', session=None):
    wf_def = get_workflow_definition(identifier, namespace)

    m_dbutils.check_db_obj_access(wf_def)

    cron_triggers = get_cron_triggers(insecure=True, workflow_id=wf_def.id)
    if cron_triggers:
        raise exc.DBError(
            "Can't delete workflow that has cron triggers associated. "
            "[workflow_identifier=%s, namespace=%s], [cron_trigger_id(s)=%s]"
            % (
                identifier,
                namespace,
                ', '.join([t.id for t in cron_triggers])
            )
        )

    event_triggers = get_event_triggers(insecure=True, workflow_id=wf_def.id)

    if event_triggers:
        raise exc.DBError(
            "Can't delete workflow that has event triggers associated. "
            "[workflow_identifier=%s], [event_trigger_id(s)=%s]" %
            (identifier, ', '.join([t.id for t in event_triggers]))
        )

    # Delete workflow members first.
    delete_resource_members(resource_type='workflow', resource_id=wf_def.id)

    session.delete(wf_def)


@b.session_aware()
def delete_workflow_definitions(session=None, **kwargs):
    return _delete_all(models.WorkflowDefinition, **kwargs)


# Code sources.

@b.session_aware()
def create_code_source(values, session=None):
    code_src = models.CodeSource()
    code_src.update(values.copy())

    try:
        code_src.save(session=session)
    except db_exc.DBDuplicateEntry:
        raise exc.DBDuplicateEntryError(
            "Duplicate entry for CodeSource ['name', 'namespace',"
            " 'project_id']: {}, {}, {}".format(code_src.name,
                                                code_src.namespace,
                                                code_src.project_id)
        )

    return code_src


@b.session_aware()
def get_code_sources(fields=None, session=None, **kwargs):
    return _get_collection(
        model=models.CodeSource,
        fields=fields,
        **kwargs
    )


@b.session_aware()
def get_code_source(identifier, fields=(), session=None, namespace=''):
    code_src = _get_db_object_by_name_and_namespace_or_id(
        models.CodeSource,
        identifier,
        namespace=namespace,
        columns=fields
    )

    if not code_src:
        raise exc.DBEntityNotFoundError(
            "Code Source not found [name=%s,namespace=%s]"
            % (identifier, namespace)
        )

    return code_src


@b.session_aware()
def load_code_source(identifier, fields=(), session=None, namespace=''):
    return _get_db_object_by_name_and_namespace_or_id(
        models.CodeSource,
        identifier,
        namespace=namespace,
        columns=fields
    )


@b.session_aware()
def update_code_source(identifier, values, namespace='', session=None):
    code_src = get_code_source(identifier, namespace=namespace)

    values['version'] = code_src.version + 1

    code_src.update(values.copy())

    return code_src


@b.session_aware()
def delete_code_source(identifier, namespace='', session=None):
    code_src = get_code_source(
        identifier,
        namespace=namespace,
        session=session
    )

    session.delete(code_src)


@b.session_aware()
def delete_code_sources(session=None, **kwargs):
    return _delete_all(models.CodeSource, **kwargs)


# Dynamic actions.

@b.session_aware()
def create_dynamic_action_definition(values, session=None):
    action_def = models.DynamicActionDefinition()

    action_def.update(values.copy())

    try:
        action_def.save(session=session)
    except db_exc.DBDuplicateEntry:
        raise exc.DBDuplicateEntryError(
            "Duplicate entry for dynamic action definition "
            "[name=%s, namespace=%s, project_id=%s]"
            % (action_def.name, action_def.namespace, action_def.project_id)
        )

    return action_def


@b.session_aware()
def update_dynamic_action_definition(identifier, values, namespace='',
                                     session=None):
    action_def = get_dynamic_action_definition(identifier, namespace=namespace)

    action_def.update(values.copy())

    return action_def


@b.session_aware()
def get_dynamic_action_definition(identifier, fields=(), namespace='',
                                  session=None):
    action = _get_db_object_by_name_and_namespace_or_id(
        models.DynamicActionDefinition,
        identifier,
        namespace=namespace,
        columns=fields
    )

    if not action:
        raise exc.DBEntityNotFoundError(
            "Dynamic Action not found [name=%s,namespace=%s]"
            % (identifier, namespace)
        )

    return action


@b.session_aware()
def load_dynamic_action_definition(identifier, fields=(), namespace='',
                                   session=None):
    return _get_db_object_by_name_and_namespace_or_id(
        models.DynamicActionDefinition,
        identifier,
        namespace=namespace,
        columns=fields
    )


@b.session_aware()
def get_dynamic_action_definitions(fields=None, session=None, **kwargs):
    return _get_collection(
        model=models.DynamicActionDefinition,
        fields=fields,
        **kwargs
    )


@b.session_aware()
def delete_dynamic_action_definition(identifier, namespace='', session=None):
    action_def = get_dynamic_action_definition(identifier, namespace)

    session.delete(action_def)


@b.session_aware()
def delete_dynamic_action_definitions(session=None, **kwargs):
    return _delete_all(models.DynamicActionDefinition, **kwargs)


# Action definitions.

@b.session_aware()
def get_action_definition_by_id(id, fields=(), session=None):
    action_def = _get_db_object_by_id(
        models.ActionDefinition,
        id,
        columns=fields
    )

    if not action_def:
        raise exc.DBEntityNotFoundError(
            "Action not found [action_id=%s]" % id
        )

    return action_def


@b.session_aware()
def get_action_definition(identifier, fields=(), session=None, namespace=''):
    a_def = _get_db_object_by_name_and_namespace_or_id(
        models.ActionDefinition,
        identifier,
        namespace=namespace,
        columns=fields
    )

    #  If the action was not found in the given namespace,
    #  look in the default namespace
    if not a_def:
        a_def = _get_db_object_by_name_and_namespace_or_id(
            models.ActionDefinition,
            identifier,
            namespace='',
            columns=fields
        )

    if not a_def:
        raise exc.DBEntityNotFoundError(
            "Action definition not found [action_name=%s, namespace=%s]"
            % (identifier, namespace)
        )

    return a_def


@b.session_aware()
def load_action_definition(name, fields=(), session=None, namespace=''):
    return _get_db_object_by_name_and_namespace(
        models.ActionDefinition,
        name,
        namespace=namespace,
        columns=fields
    )


@b.session_aware()
def get_action_definitions(session=None, **kwargs):
    return _get_collection(model=models.ActionDefinition, **kwargs)


@b.session_aware()
def create_action_definition(values, session=None):
    a_def = models.ActionDefinition()

    a_def.update(values)

    try:
        a_def.save(session=session)
    except db_exc.DBDuplicateEntry:
        raise exc.DBDuplicateEntryError(
            "Duplicate entry for Action ['name', 'namespace', 'project_id']:"
            " {}, {}, {}".format(a_def.name, a_def.namespace, a_def.project_id)
        )

    return a_def


@b.session_aware()
def update_action_definition(identifier, values, session=None):
    namespace = values.get('namespace', '')
    a_def = get_action_definition(identifier, namespace=namespace)

    a_def.update(values.copy())

    return a_def


@b.session_aware()
def create_or_update_action_definition(name, values, session=None):
    namespace = values.get('namespace', '')
    if not _get_db_object_by_name_and_namespace(
            models.ActionDefinition,
            name,
            namespace=namespace):
        return create_action_definition(values)
    else:
        return update_action_definition(name, values)


@b.session_aware()
def delete_action_definition(identifier, namespace='', session=None):
    a_def = get_action_definition(identifier, namespace=namespace)

    session.delete(a_def)


@b.session_aware()
def delete_action_definitions(session=None, **kwargs):
    return _delete_all(models.ActionDefinition, **kwargs)


# Action executions.

@b.session_aware()
def get_action_execution(id, insecure=False, fields=(), session=None):
    a_ex = _get_db_object_by_id(models.ActionExecution, id, insecure=insecure,
                                columns=fields)

    if not a_ex:
        raise exc.DBEntityNotFoundError(
            "ActionExecution not found [id=%s]" % id
        )

    return a_ex


@b.session_aware()
def load_action_execution(id, fields=(), session=None):
    return _get_db_object_by_id(models.ActionExecution, id, columns=fields)


@b.session_aware()
def get_action_executions(session=None, **kwargs):
    return _get_action_executions(**kwargs)


@b.session_aware()
def create_action_execution(values, session=None):
    a_ex = models.ActionExecution()

    a_ex.update(values.copy())

    try:
        a_ex.save(session=session)
    except db_exc.DBDuplicateEntry as e:
        raise exc.DBDuplicateEntryError(
            "Duplicate entry for ActionExecution ID: {}".format(e.value)
        )

    return a_ex


@b.session_aware()
def update_action_execution(id, values, insecure=False, session=None):
    a_ex = get_action_execution(id, insecure)

    a_ex.update(values.copy())

    return a_ex


@b.session_aware()
def create_or_update_action_execution(id, values, session=None):
    if not _get_db_object_by_id(models.ActionExecution, id):
        return create_action_execution(values)
    else:
        return update_action_execution(id, values)


@b.session_aware()
def update_action_execution_heartbeat(id, session=None):
    if not id:
        raise exc.DBEntityNotFoundError

    now = utils.utc_now_sec()
    session.query(models.ActionExecution). \
        filter(models.ActionExecution.id == id). \
        update({'last_heartbeat': now})


@b.session_aware()
def delete_action_execution(id, session=None):
    count = _secure_query(models.ActionExecution).filter(
        models.ActionExecution.id == id).delete()

    if count == 0:
        raise exc.DBEntityNotFoundError(
            "ActionExecution not found [id=%s]" % id
        )


@b.session_aware()
def delete_action_executions(session=None, **kwargs):
    return _delete_all(models.ActionExecution, **kwargs)


def _get_action_executions(**kwargs):
    return _get_collection(models.ActionExecution, **kwargs)


# Workflow executions.

@b.session_aware()
def get_workflow_execution(id, fields=(), session=None):
    ctx = context.ctx()

    wf_ex = _get_db_object_by_id(
        models.WorkflowExecution,
        id,
        insecure=ctx.is_admin,
        columns=fields
    )

    if not wf_ex:
        raise exc.DBEntityNotFoundError(
            "WorkflowExecution not found [id=%s]" % id
        )

    return wf_ex


@b.session_aware()
def load_workflow_execution(id, fields=(), session=None):
    return _get_db_object_by_id(models.WorkflowExecution, id, columns=fields)


@b.session_aware()
def get_workflow_executions(session=None, **kwargs):
    return _get_collection(models.WorkflowExecution, **kwargs)


@b.session_aware()
def create_workflow_execution(values, session=None):
    wf_ex = models.WorkflowExecution()

    wf_ex.update(values.copy())

    try:
        wf_ex.save(session=session)
    except db_exc.DBDuplicateEntry as e:
        raise exc.DBDuplicateEntryError(
            "Duplicate entry for WorkflowExecution with ID: {value} ".format(
                value=e.value
            )
        )

    return wf_ex


@b.session_aware()
def update_workflow_execution(id, values, session=None):
    wf_ex = get_workflow_execution(id)

    m_dbutils.check_db_obj_access(wf_ex)

    wf_ex.update(values.copy())

    return wf_ex


@b.session_aware()
def create_or_update_workflow_execution(id, values, session=None):
    if not _get_db_object_by_id(models.WorkflowExecution, id):
        return create_workflow_execution(values)
    else:
        return update_workflow_execution(id, values)


@b.session_aware()
def delete_workflow_execution(id, session=None):
    model = models.WorkflowExecution
    insecure = context.ctx().is_admin
    query = b.model_query(model) if insecure else _secure_query(model)

    try:
        count = query.filter(
            models.WorkflowExecution.id == id
        ).delete()

        if count == 0:
            raise exc.DBEntityNotFoundError(
                "WorkflowExecution not found [id=%s]" % id
            )
    except Exception as e:
        if is_mysql_max_depth_error(e) or is_mariadb_max_depth_error(e):
            # https://bugs.launchpad.net/mistral/+bug/1832300
            # mysql cascade delete error
            delete_workflow_execution_recurse(id)
        else:
            raise


def is_mysql_max_depth_error(e):
    pattern = ".*3008.*Foreign key cascade delete" \
              "/update exceeds max depth of 15.*"
    return re.match(pattern, str(e))


def is_mariadb_max_depth_error(e):
    pattern = ".*Got error 193.*ON DELETE CASCADE.*"
    return re.match(pattern, str(e))


def delete_workflow_execution_recurse(wf_ex_id):
    sub_wf_ex_ids = _get_all_direct_subworkflows(wf_ex_id)

    for sub_wf_ex_id in sub_wf_ex_ids:
        delete_workflow_execution(sub_wf_ex_id)

    delete_workflow_execution(wf_ex_id)


def _get_all_direct_subworkflows(wf_ex_id):
    model = models.WorkflowExecution
    insecure = context.ctx().is_admin
    if insecure:
        query = b.model_query(model, [model.id])
    else:
        query = _secure_query(model, model.id)
    query = query.join(
        models.TaskExecution,
        models.WorkflowExecution.task_execution_id == models.TaskExecution.id
    ).filter(
        models.TaskExecution.workflow_execution_id == wf_ex_id
    )

    return [i[0] for i in query.all()]


@b.session_aware()
def delete_workflow_executions(session=None, **kwargs):
    return _delete_all(models.WorkflowExecution, **kwargs)


def update_workflow_execution_state(id, cur_state, state):
    specimen = models.WorkflowExecution(id=id, state=cur_state)

    return update_on_match(id, specimen, values={'state': state}, attempts=1)


# Tasks executions.

@b.session_aware()
def get_task_execution(id, fields=(), session=None):
    task_ex = _get_db_object_by_id(models.TaskExecution, id, columns=fields)

    if not task_ex:
        raise exc.DBEntityNotFoundError(
            "Task execution not found [id=%s]" % id
        )

    return task_ex


@b.session_aware()
def load_task_execution(id, fields=(), session=None):
    return _get_db_object_by_id(models.TaskExecution, id, columns=fields)


@b.session_aware()
def get_task_executions(session=None, **kwargs):
    return _get_collection(models.TaskExecution, **kwargs)


@b.session_aware()
def get_task_executions_count(session=None, **kwargs):
    query = b.model_query(models.TaskExecution)

    query = query.filter_by(**kwargs)

    return query.count()


def _get_completed_task_executions_query(kwargs):
    query = b.model_query(models.TaskExecution)

    query = query.filter_by(**kwargs)

    query = query.filter(
        models.TaskExecution.state.in_(
            [states.ERROR,
             states.CANCELLED,
             states.SKIPPED,
             states.SUCCESS]
        )
    )

    return query


@b.session_aware()
def get_completed_task_executions(session=None, **kwargs):
    query = _get_completed_task_executions_query(kwargs)

    return query.all()


@b.session_aware()
def get_completed_task_executions_as_batches(session=None, **kwargs):
    # NOTE: Using batch querying seriously allows to optimize memory
    # consumption on operations when we need to iterate through
    # a list of task executions and do some processing like merging
    # their inbound contexts. If we don't use batches Mistral has to
    # hold all the collection (that can be large) in memory.
    # Using a generator that returns batches lets GC to collect a
    # batch of task executions that has already been processed.
    query = _get_completed_task_executions_query(kwargs)

    # Batch size 20 may be arguable but still seems reasonable: it's big
    # enough to keep the total number of DB hops small (say for 100 tasks
    # we'll need only 5) and small enough not to drastically increase
    # memory footprint if the number of tasks is big like several hundreds.
    batch_size = 20
    idx = 0

    while idx < query.count():
        yield query.slice(idx, idx + batch_size).all()

        idx += batch_size


def _get_incomplete_task_executions_query(kwargs):
    query = b.model_query(models.TaskExecution)

    query = query.filter_by(**kwargs)

    query = query.filter(
        models.TaskExecution.state.in_(
            [states.IDLE,
             states.RUNNING,
             states.WAITING,
             states.RUNNING_DELAYED,
             states.PAUSED]
        )
    )

    return query


@b.session_aware()
def get_incomplete_task_executions(session=None, **kwargs):
    query = _get_incomplete_task_executions_query(kwargs)

    return query.all()


@b.session_aware()
def get_incomplete_task_executions_count(session=None, **kwargs):
    query = _get_incomplete_task_executions_query(kwargs)

    return query.count()


@b.session_aware()
def create_task_execution(values, session=None):
    task_ex = models.TaskExecution()

    task_ex.update(values)

    try:
        task_ex.save(session=session)
    except db_exc.DBDuplicateEntry as e:
        raise exc.DBDuplicateEntryError(
            "Duplicate entry for TaskExecution ID: {}".format(e.value)
        )

    return task_ex


@b.session_aware()
def update_task_execution(id, values, session=None):
    task_ex = get_task_execution(id)

    task_ex.update(values.copy())

    return task_ex


@b.session_aware()
def create_or_update_task_execution(id, values, session=None):
    if not _get_db_object_by_id(models.TaskExecution, id):
        return create_task_execution(values)
    else:
        return update_task_execution(id, values)


@b.session_aware()
def delete_task_execution(id, session=None):
    count = _secure_query(models.TaskExecution).filter(
        models.TaskExecution.id == id).delete()

    if count == 0:
        raise exc.DBEntityNotFoundError(
            "Task execution not found [id=%s]" % id
        )


@b.session_aware()
def delete_task_executions(session=None, **kwargs):
    return _delete_all(models.TaskExecution, **kwargs)


def update_task_execution_state(id, cur_state, state):
    specimen = models.TaskExecution(id=id, state=cur_state)

    return update_on_match(id, specimen, values={'state': state}, attempts=1)


# Delayed calls.

@b.session_aware()
def create_delayed_call(values, session=None):
    delayed_call = models.DelayedCall()
    delayed_call.update(values.copy())

    try:
        delayed_call.save(session)
    except db_exc.DBDuplicateEntry as e:
        raise exc.DBDuplicateEntryError(
            "Duplicate entry for DelayedCall ID: {}".format(e.value)
        )

    return delayed_call


@b.session_aware()
def delete_delayed_call(id, session=None):
    # It's safe to use insecure query here because users can't access
    # delayed calls.
    count = b.model_query(models.DelayedCall).filter(
        models.DelayedCall.id == id).delete()

    if count == 0:
        raise exc.DBEntityNotFoundError(
            "Delayed Call not found [id=%s]" % id
        )


@b.session_aware()
def get_delayed_calls_to_start(time, batch_size=None, session=None):
    query = b.model_query(models.DelayedCall)

    query = query.filter(models.DelayedCall.execution_time < time)
    query = query.filter_by(processing=False)
    query = query.order_by(models.DelayedCall.execution_time)
    query = query.limit(batch_size)

    return query.all()


@b.session_aware()
def update_delayed_call(id, values, query_filter=None, session=None):
    if query_filter:
        try:
            specimen = models.DelayedCall(id=id, **query_filter)
            delayed_call = b.model_query(
                models.DelayedCall).update_on_match(specimen=specimen,
                                                    surrogate_key='id',
                                                    values=values)
            return delayed_call, 1

        except oslo_sqlalchemy.update_match.NoRowsMatched as e:
            LOG.debug(
                "No rows matched for update call [id=%s, values=%s, "
                "query_filter=%s,"
                "exception=%s]", id, values, query_filter, e
            )

            return None, 0

    else:
        delayed_call = get_delayed_call(id=id, session=session)
        delayed_call.update(values)

        return delayed_call, len(session.dirty)


@b.session_aware()
def get_delayed_call(id, session=None):
    delayed_call = _get_db_object_by_id(models.DelayedCall, id)

    if not delayed_call:
        raise exc.DBEntityNotFoundError(
            "Delayed Call not found [id=%s]" % id
        )

    return delayed_call


@b.session_aware()
def get_delayed_calls(session=None, **kwargs):
    return _get_collection(model=models.DelayedCall, **kwargs)


@b.session_aware()
def get_delayed_calls_count(session=None, **kwargs):
    return _get_count(model=models.DelayedCall, **kwargs)


@b.session_aware()
def delete_delayed_calls(session=None, **kwargs):
    return _delete_all(models.DelayedCall, **kwargs)


# Scheduled jobs.

@b.session_aware()
def create_scheduled_job(values, session=None):
    job = models.ScheduledJob()

    job.update(values.copy())

    try:
        job.save(session)
    except db_exc.DBDuplicateEntry as e:
        raise exc.DBDuplicateEntryError(
            "Duplicate entry for ScheduledJob ID: {}".format(e.value)
        )

    return job


@b.session_aware()
def get_scheduled_jobs_to_start(time, batch_size=None, session=None):
    query = b.model_query(models.ScheduledJob)

    execute_at_col = models.ScheduledJob.execute_at
    captured_at_col = models.ScheduledJob.captured_at

    # Filter by execution time accounting for a configured job pickup interval.
    # TODO(rakhmerov): Configuration options should not be accessed here.
    query = query.filter(
        execute_at_col <
        time - datetime.timedelta(seconds=CONF.scheduler.pickup_job_after)
    )

    # Filter by captured time accounting for a configured captured job timeout.
    min_captured_at = (
        utils.utc_now_sec() -
        datetime.timedelta(seconds=CONF.scheduler.captured_job_timeout)
    )

    query = query.filter(
        sa.or_(
            captured_at_col == sa.null(),
            captured_at_col <= min_captured_at
        )
    )

    query = query.order_by(execute_at_col)
    query = query.limit(batch_size)

    return query.all()


@b.session_aware()
def update_scheduled_job(id, values, query_filter=None, session=None):
    if query_filter:
        try:
            specimen = models.ScheduledJob(id=id, **query_filter)

            job = b.model_query(
                models.ScheduledJob
            ).update_on_match(
                specimen=specimen,
                surrogate_key='id',
                values=values
            )

            return job, 1

        except oslo_sqlalchemy.update_match.NoRowsMatched as e:
            LOG.debug(
                "No rows matched for update scheduled job [id=%s, values=%s, "
                "query_filter=%s,"
                "exception=%s]", id, values, query_filter, e
            )

            return None, 0

    else:
        job = get_scheduled_job(id=id, session=session)

        job.update(values)

        return job, len(session.dirty)


@b.session_aware()
def get_scheduled_job(id, session=None):
    job = _get_db_object_by_id(models.ScheduledJob, id)

    if not job:
        raise exc.DBEntityNotFoundError(
            "Scheduled job not found [id=%s]" % id
        )

    return job


@b.session_aware()
def delete_scheduled_job(id, session=None):
    # It's safe to use insecure query here because users can't access
    # scheduled job.
    count = b.model_query(models.ScheduledJob).filter(
        models.ScheduledJob.id == id).delete()

    if count == 0:
        raise exc.DBEntityNotFoundError(
            "Scheduled job not found [id=%s]" % id
        )


def get_scheduled_jobs(**kwargs):
    return _get_collection(model=models.ScheduledJob, **kwargs)


@b.session_aware()
def delete_scheduled_jobs(session=None, **kwargs):
    return _delete_all(models.ScheduledJob, **kwargs)


def get_scheduled_jobs_count(**kwargs):
    return _get_count(model=models.ScheduledJob, **kwargs)


# Other functions.

@b.session_aware()
def get_expired_executions(expiration_time, limit=None, columns=(),
                           session=None):
    query = _get_completed_root_executions_query(columns)
    query = query.filter(models.WorkflowExecution.updated_at < expiration_time)

    if limit:
        query = query.limit(limit)

    return query.all()


@b.session_aware()
def get_running_expired_sync_action_executions(expiration_time,
                                               limit, session=None):
    query = b.model_query(models.ActionExecution)

    query = query.filter(
        models.ActionExecution.last_heartbeat < expiration_time
    )
    query = query.filter_by(is_sync=True)
    query = query.filter(models.ActionExecution.state == states.RUNNING)

    if limit:
        query.limit(limit)

    return query.all()


@b.session_aware()
def get_superfluous_executions(max_finished_executions, limit=None, columns=(),
                               session=None):
    if not max_finished_executions:
        return []

    query = _get_completed_root_executions_query(columns)
    query = query.order_by(models.WorkflowExecution.updated_at.desc())
    query = query.offset(max_finished_executions)

    if limit:
        query = query.limit(limit)

    return query.all()


def _get_completed_root_executions_query(columns):
    query = b.model_query(models.WorkflowExecution, columns=columns)

    # This is an empty list by default.
    ignored_states = CONF.execution_expiration_policy.ignored_states
    desired_states = states.TERMINAL_STATES - set(ignored_states)

    # Only workflow executions that are not a child of
    # other workflow executions.
    query = query.filter(
        models.WorkflowExecution.task_execution_id == sa.null()
    )

    query = query.filter(
        models.WorkflowExecution.state.in_(desired_states)
    )

    return query


@b.session_aware()
def get_cron_trigger(identifier, session=None, fields=()):
    ctx = context.ctx()

    cron_trigger = _get_db_object_by_name_and_namespace_or_id(
        models.CronTrigger,
        identifier,
        insecure=ctx.is_admin,
        columns=fields,
    )

    if not cron_trigger:
        raise exc.DBEntityNotFoundError(
            "Cron trigger not found [identifier=%s]" % identifier
        )

    return cron_trigger


@b.session_aware()
def get_cron_trigger_by_id(id, session=None, fields=()):
    ctx = context.ctx()
    cron_trigger = _get_db_object_by_id(models.CronTrigger, id,
                                        insecure=ctx.is_admin,
                                        columns=fields)
    if not cron_trigger:
        raise exc.DBEntityNotFoundError(
            "Cron trigger not found [id=%s]" % id
        )

    return cron_trigger


@b.session_aware()
def load_cron_trigger(identifier, session=None, fields=()):
    return _get_db_object_by_name_and_namespace_or_id(
        models.CronTrigger,
        identifier,
        columns=fields,
    )


@b.session_aware()
def get_cron_triggers(session=None, **kwargs):
    return _get_collection(models.CronTrigger, **kwargs)


@b.session_aware()
def get_next_cron_triggers(time, session=None):
    query = b.model_query(models.CronTrigger)

    query = query.filter(models.CronTrigger.next_execution_time < time)
    query = query.order_by(models.CronTrigger.next_execution_time)

    return query.all()


@b.session_aware()
def create_cron_trigger(values, session=None):
    cron_trigger = models.CronTrigger()

    cron_trigger.update(values)

    try:
        cron_trigger.save(session=session)
    except db_exc.DBDuplicateEntry:
        raise exc.DBDuplicateEntryError(
            "Duplicate entry for cron trigger ['name', 'project_id']: "
            "{}, {}".format(cron_trigger.name, cron_trigger.project_id)
        )
    # TODO(nmakhotkin): Remove this 'except' after fixing
    # https://bugs.launchpad.net/oslo.db/+bug/1458583.
    except db_exc.DBError as e:
        raise exc.DBDuplicateEntryError(
            "Duplicate entry for cron trigger: %s" % e
        )

    return cron_trigger


@b.session_aware()
def update_cron_trigger(identifier, values, session=None, query_filter=None):
    cron_trigger = get_cron_trigger(identifier)

    if query_filter:
        try:
            # Execute the UPDATE statement with the query_filter as the WHERE.
            specimen = models.CronTrigger(id=cron_trigger.id, **query_filter)

            query = b.model_query(models.CronTrigger)

            cron_trigger = query.update_on_match(
                specimen=specimen,
                surrogate_key='id',
                values=values
            )

            return cron_trigger, 1

        except oslo_sqlalchemy.update_match.NoRowsMatched:
            LOG.debug(
                "No rows matched for cron update call"
                "[id=%s, values=%s, query_filter=%s", id, values, query_filter
            )

            return cron_trigger, 0

    else:
        cron_trigger.update(values.copy())

        return cron_trigger, len(session.dirty)


@b.session_aware()
def create_or_update_cron_trigger(identifier, values, session=None):
    cron_trigger = _get_db_object_by_name_and_namespace_or_id(
        models.CronTrigger,
        identifier
    )

    if not cron_trigger:
        return create_cron_trigger(values)
    else:
        updated, _ = update_cron_trigger(identifier, values)
        return updated


@b.session_aware()
def delete_cron_trigger(identifier, session=None):
    cron_trigger = get_cron_trigger(identifier)

    m_dbutils.check_db_obj_access(cron_trigger)
    # Delete the cron trigger by ID and get the affected row count.
    table = models.CronTrigger.__table__
    result = session.execute(
        table.delete().where(table.c.id == cron_trigger.id)
    )

    return result.rowcount


@b.session_aware()
def delete_cron_triggers(session=None, **kwargs):
    return _delete_all(models.CronTrigger, **kwargs)


# Environments.

@b.session_aware()
def get_environment(name, session=None, fields=()):
    env = _get_db_object_by_name(models.Environment, name, columns=fields)

    if not env:
        raise exc.DBEntityNotFoundError(
            "Environment not found [name=%s]" % name
        )

    return env


@b.session_aware()
def load_environment(name, session=None, fields=()):
    return _get_db_object_by_name(models.Environment, name, columns=fields)


@b.session_aware()
def get_environments(session=None, **kwargs):
    return _get_collection(models.Environment, **kwargs)


@b.session_aware()
def create_environment(values, session=None):
    env = models.Environment()

    env.update(values)

    try:
        env.save(session=session)
    except db_exc.DBDuplicateEntry:
        raise exc.DBDuplicateEntryError(
            "Duplicate entry for Environment ['name', 'project_id']:"
            " {}, {}".format(env.name, env.project_id)
        )

    return env


@b.session_aware()
def update_environment(name, values, session=None):
    env = get_environment(name)

    env.update(values)

    return env


@b.session_aware()
def create_or_update_environment(name, values, session=None):
    env = _get_db_object_by_name(models.Environment, name)

    if not env:
        return create_environment(values)
    else:
        return update_environment(name, values)


@b.session_aware()
def delete_environment(name, session=None):
    count = _secure_query(models.Environment).filter(
        models.Environment.name == name).delete()

    if count == 0:
        raise exc.DBEntityNotFoundError(
            "Environment not found [name=%s]" % name
        )


@b.session_aware()
def delete_environments(session=None, **kwargs):
    return _delete_all(models.Environment, **kwargs)


# Resource members.


RESOURCE_MAPPING = {
    models.WorkflowDefinition: 'workflow',
    models.Workbook: 'workbook'
}


def _get_criterion(resource_id, member_id=None, is_owner=True):
    """Generates criterion for querying resource_member_v2 table."""

    # Resource owner query resource membership with member_id.
    if is_owner and member_id:
        return sa.and_(
            models.ResourceMember.project_id == security.get_project_id(),
            models.ResourceMember.resource_id == resource_id,
            models.ResourceMember.member_id == member_id
        )
    # Resource owner query resource memberships.
    elif is_owner and not member_id:
        return sa.and_(
            models.ResourceMember.project_id == security.get_project_id(),
            models.ResourceMember.resource_id == resource_id,
        )

    # Other members query other resource membership.
    elif not is_owner and member_id and member_id != security.get_project_id():
        return None

    # Resource member query resource memberships.
    return sa.and_(
        models.ResourceMember.member_id == security.get_project_id(),
        models.ResourceMember.resource_id == resource_id
    )


@b.session_aware()
def create_resource_member(values, session=None):
    res_member = models.ResourceMember()

    res_member.update(values.copy())

    try:
        res_member.save(session=session)
    except db_exc.DBDuplicateEntry:
        raise exc.DBDuplicateEntryError(
            "Duplicate entry for ResourceMember ['resource_id',"
            " 'resource_type', 'member_id']: {}, {}, "
            "{}".format(res_member.resource_id,
                        res_member.resource_type,
                        res_member.member_id)
        )

    return res_member


@b.session_aware()
def get_resource_member(resource_id, res_type, member_id, session=None):
    query = _secure_query(models.ResourceMember).filter_by(
        resource_type=res_type
    )

    # Both resource owner and resource member can do query.
    res_member = query.filter(
        sa.or_(
            _get_criterion(resource_id, member_id),
            _get_criterion(resource_id, member_id, is_owner=False)
        )
    ).first()

    if not res_member:
        raise exc.DBEntityNotFoundError(
            "Resource member not found [resource_id=%s, member_id=%s]" %
            (resource_id, member_id)
        )

    return res_member


@b.session_aware()
def get_resource_members(resource_id, res_type, session=None):
    query = _secure_query(models.ResourceMember).filter_by(
        resource_type=res_type
    )

    # Both resource owner and resource member can do query.
    res_members = query.filter(
        sa.or_(
            _get_criterion(resource_id),
            _get_criterion(resource_id, is_owner=False),
        )
    ).all()

    return res_members


@b.session_aware()
def update_resource_member(resource_id, res_type, member_id, values,
                           session=None):
    # Only member who is not the owner of the resource can update the
    # membership status.
    if member_id != security.get_project_id():
        raise exc.DBEntityNotFoundError(
            "Resource member not found [resource_id=%s, member_id=%s]" %
            (resource_id, member_id)
        )

    query = _secure_query(models.ResourceMember).filter_by(
        resource_type=res_type
    )

    res_member = query.filter(
        _get_criterion(resource_id, member_id, is_owner=False)
    ).first()

    if not res_member:
        raise exc.DBEntityNotFoundError(
            "Resource member not found [resource_id=%s, member_id=%s]" %
            (resource_id, member_id)
        )

    res_member.update(values.copy())

    return res_member


@b.session_aware()
def delete_resource_member(resource_id, res_type, member_id, session=None):
    query = _secure_query(models.ResourceMember). \
        filter_by(resource_type=res_type). \
        filter(_get_criterion(resource_id, member_id))

    # TODO(kong): Check association with cron triggers when deleting a workflow
    # member which is in 'accepted' status.

    count = query.delete()

    if count == 0:
        raise exc.DBEntityNotFoundError(
            "Resource member not found [resource_id=%s, member_id=%s]" %
            (resource_id, member_id)
        )


@b.session_aware()
def delete_resource_members(session=None, **kwargs):
    return _delete_all(models.ResourceMember, **kwargs)


def _get_accepted_resources(res_type):
    resources = _secure_query(models.ResourceMember).filter(
        sa.and_(
            models.ResourceMember.resource_type == res_type,
            models.ResourceMember.status == 'accepted',
            models.ResourceMember.member_id == security.get_project_id()
        )
    ).all()

    return resources


# Event triggers.

@b.session_aware()
def get_event_trigger(id, insecure=False, session=None, fields=()):
    event_trigger = _get_db_object_by_id(models.EventTrigger, id, insecure,
                                         columns=fields)

    if not event_trigger:
        raise exc.DBEntityNotFoundError(
            "Event trigger not found [id=%s]." % id
        )

    return event_trigger


@b.session_aware()
def load_event_trigger(id, insecure=False, session=None, fields=()):
    return _get_db_object_by_id(models.EventTrigger, id, insecure,
                                columns=fields)


@b.session_aware()
def get_event_triggers(session=None, **kwargs):
    return _get_collection(model=models.EventTrigger, **kwargs)


@b.session_aware()
def create_event_trigger(values, session=None):
    event_trigger = models.EventTrigger()

    event_trigger.update(values)

    try:
        event_trigger.save(session=session)
    except db_exc.DBDuplicateEntry:
        raise exc.DBDuplicateEntryError(
            "Duplicate entry for EventTrigger ['exchange', 'topic',"
            " 'event', 'workflow_id', 'project_id']:"
            " {}, {}, {}, {}, {}".format(event_trigger.exchange,
                                         event_trigger.topic,
                                         event_trigger.event,
                                         event_trigger.workflow_id,
                                         event_trigger.project_id))
    # TODO(nmakhotkin): Remove this 'except' after fixing
    # https://bugs.launchpad.net/oslo.db/+bug/1458583.
    except db_exc.DBError as e:
        raise exc.DBDuplicateEntryError(
            "Duplicate entry for event trigger: %s" % e
        )

    return event_trigger


@b.session_aware()
def update_event_trigger(id, values, session=None):
    event_trigger = get_event_trigger(id)

    event_trigger.update(values.copy())

    return event_trigger


@b.session_aware()
def delete_event_trigger(id, session=None):
    # It's safe to use insecure query here because users can't access
    # delayed calls.
    count = b.model_query(models.EventTrigger).filter(
        models.EventTrigger.id == id).delete()

    if count == 0:
        raise exc.DBEntityNotFoundError(
            "Event trigger not found [id=%s]." % id
        )


@b.session_aware()
def delete_event_triggers(session=None, **kwargs):
    return _delete_all(models.EventTrigger, **kwargs)


# Locks.

@b.session_aware()
def create_named_lock(name, session=None):
    # This method has to work not through SQLAlchemy session because
    # session may not immediately issue an SQL query to a database
    # and instead just schedule it whereas we need to make sure to
    # issue a query immediately.
    session.flush()

    insert = models.NamedLock.__table__.insert()

    lock_id = utils.generate_unicode_uuid()

    session.execute(insert.values(id=lock_id, name=name))

    session.flush()

    return lock_id


@b.session_aware()
def get_named_locks(session=None, **kwargs):
    return _get_collection(models.NamedLock, **kwargs)


@b.session_aware()
def delete_named_lock(lock_id, session=None):
    # This method has to work without SQLAlchemy session because
    # session may not immediately issue an SQL query to a database
    # and instead just schedule it whereas we need to make sure to
    # issue a query immediately.
    session.flush()

    table = models.NamedLock.__table__

    delete = table.delete()

    session.execute(delete.where(table.c.id == lock_id))

    session.flush()


@contextlib.contextmanager
def named_lock(name):
    # NOTE(rakhmerov): We can't use the well-known try-finally pattern here
    # because if lock creation failed then it means that the SQLAlchemy
    # session is no longer valid and we can't use to try to delete the lock.
    # All we can do here is to let the exception bubble up so that the
    # transaction management code could rollback the transaction.

    lock_id = create_named_lock(name)

    yield

    delete_named_lock(lock_id)


def get_maintenance_status():
    query = b.get_engine().execute('SELECT status FROM mistral_metrics '
                                   'WHERE name = %s', ("maintenance_status",))
    rows = query.fetchall()

    return rows[0][0] if rows else None


def update_maintenance_status(status):
    b.get_engine().execute(
        "INSERT into mistral_metrics (id, name, value)"
        "VALUES (1, %s, %s)"
        "ON CONFLICT (name) DO UPDATE"
        "SET value = EXCLUDED.value",
        ("maintenance_status", status,))


@b.session_aware()
def get_overdue_calls(time, session=None):
    query = b.model_query(models.DelayedCall)

    query = query.filter(models.DelayedCall.execution_time < time)
    query = query.filter_by(processing=True)

    return query.all()



# Copyright 2014 - Mirantis, Inc.
# Copyright 2014 - StackStorm, Inc.
# Copyright 2022 - NetCracker Technology Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from email import header
from email.mime import multipart
from email.mime import text
import smtplib
import time
from urllib import parse

from oslo_log import log as logging
import requests

from mistral import exceptions as exc
from mistral import utils
from mistral.utils import javascript
from mistral.utils import rest_utils
from mistral.utils import ssh_utils
from mistral_lib import actions

LOG = logging.getLogger(__name__)


class EchoAction(actions.Action):
    """Echo action.

    This action just returns a configured value as a result without doing
    anything else. The value of such action implementation is that it
    can be used in development (for testing), demonstration and designing
    of workflows themselves where echo action can play the role of temporary
    stub.
    """

    def __init__(self, output, delay=0):
        super(EchoAction, self).__init__()

        self.output = output

        try:
            self._delay = float(delay)
            self._delay = 0 if self._delay < 0 else self._delay
        except ValueError:
            self._delay = 0

    def run(self, context):
        LOG.info(
            'Running echo action [output=%s, delay=%s]',
            self.output,
            self._delay
        )

        time.sleep(self._delay)

        return self.output

    def test(self, context):
        return 'Echo'

    def __eq__(self, other):
        if type(other) is not EchoAction:
            return False

        if self.output != other.output or self._delay != other._delay:
            return False

        return True


class NoOpAction(actions.Action):
    """No-operation action.

    This action does nothing. It can be mostly useful for testing and
    debugging purposes.
    """
    def run(self, context):
        LOG.info('Running no-op action')

        return None

    def test(self, context):
        return None


class AsyncNoOpAction(NoOpAction):
    """Asynchronous no-operation action."""
    def is_sync(self):
        return False


class FailAction(actions.Action):
    """'Always fail' action.

    If you pass the `error_data` parameter, this action will be failed and
    return this data as error data. Otherwise, the action just throws an
    instance of ActionException.

    This behavior is useful in a number of cases, especially if we need to
    test a scenario where some of workflow tasks fail.

    :param error_data: Action will be failed with this data
    """

    def __init__(self, error_data=None):
        super(FailAction, self).__init__()

        self.error_data = error_data

    def __repr__(self):
        return 'Fail action'

    def run(self, context):
        LOG.info('Running fail action.')

        if self.error_data:
            return actions.Result(error=self.error_data)

        raise exc.ActionException('Fail action expected exception.')

    def test(self, context):
        if self.error_data:
            return actions.Result(error=self.error_data)

        raise exc.ActionException('Fail action expected exception.')


class HTTPAction(actions.Action):
    """HTTP action.

    :param url: URL for the new HTTP request.
    :param method: (optional, 'GET' by default) method for the new HTTP
        request.
    :param params: (optional) Dictionary or bytes to be sent in the
        query string for the HTTP request.
    :param body: (optional) Dictionary, bytes, or file-like object to send
        in the body of the HTTP request.
    :param json: (optional) A JSON serializable Python object to send
        in the body of the HTTP request.
    :param headers: (optional) Dictionary of HTTP Headers to send with
        the HTTP request.
    :param cookies: (optional) Dict or CookieJar object to send with
        the HTTP request.
    :param auth: (optional) Auth tuple to enable Basic/Digest/Custom
        HTTP Auth.
    :param timeout: (optional) Float describing the timeout of the request
        in seconds.
    :param allow_redirects: (optional) Boolean. Set to True if POST/PUT/DELETE
        redirect following is allowed.
    :param proxies: (optional) Dictionary mapping protocol to the URL of
        the proxy.
    :param verify: (optional) if ``True``, the SSL cert will be verified.
        A CA_BUNDLE path can also be provided.
    """

    def __init__(self,
                 url,
                 method="GET",
                 params=None,
                 body=None,
                 json=None,
                 headers=None,
                 cookies=None,
                 auth=None,
                 timeout=None,
                 allow_redirects=None,
                 proxies=None,
                 verify=None):
        super(HTTPAction, self).__init__()

        if auth and len(auth.split(':')) == 2:
            self.auth = (auth.split(':')[0], auth.split(':')[1])
        else:
            self.auth = auth

        if isinstance(headers, dict):
            for key, val in headers.items():
                if isinstance(val, (int, float)):
                    headers[key] = str(val)

        if body and json:
            raise exc.ActionException(
                "Only one of the parameters 'json' and 'body' can be passed"
            )

        self.url = url
        self.method = method
        self.params = params
        self.body = utils.to_json_str(body) if isinstance(body, dict) else body
        self.json = json
        self.headers = headers
        self.cookies = cookies
        self.timeout = timeout
        self.allow_redirects = allow_redirects
        self.proxies = proxies
        self.verify = verify

    def run(self, context):
        LOG.info(
            "Running HTTP action "
            "[url=%s, method=%s, params=%s, body=%s, json=%s,"
            " headers=%s, cookies=%s, auth=%s, timeout=%s,"
            " allow_redirects=%s, proxies=%s, verify=%s]",
            self.url,
            self.method,
            self.params,
            rest_utils.prepare_request_body_log(self.body),
            self.json,
            rest_utils.clear_sensitive_headers(self.headers),
            self.cookies,
            self.auth,
            self.timeout,
            self.allow_redirects,
            self.proxies,
            self.verify
        )

        try:
            url_data = parse.urlsplit(self.url)
            if 'https' == url_data.scheme:
                action_verify = self.verify
            else:
                action_verify = None

            resp = requests.request(
                self.method,
                self.url,
                params=self.params,
                data=self.body,
                json=self.json,
                headers=self.headers,
                cookies=self.cookies,
                auth=self.auth,
                timeout=self.timeout,
                allow_redirects=self.allow_redirects,
                proxies=self.proxies,
                verify=action_verify
            )
        except Exception as e:
            LOG.exception(
                "Failed to send HTTP request for action execution: %s",
                context.execution.action_execution_id
            )
            raise exc.ActionException("Failed to send HTTP request: %s" % e)

        LOG.info(
            "HTTP action response:\n%s\n%s",
            resp.status_code,
            rest_utils.prepare_response_body_log(resp.content)
        )

        # Represent important resp data as a dictionary.
        try:
            content = resp.json()
        except Exception:
            LOG.debug("HTTP action response is not json.")
            content = resp.content
            if content and resp.encoding not in (None, 'utf-8'):
                content = content.decode(resp.encoding).encode('utf-8')

        _result = {
            'content': content,
            'status': resp.status_code,
            'headers': dict(resp.headers.items()),
            'url': resp.url,
            'history': resp.history,
            'encoding': resp.encoding,
            'reason': resp.reason,
            'cookies': dict(resp.cookies.items()),
            'elapsed': resp.elapsed.total_seconds()
        }

        if resp.status_code not in range(200, 307):
            return actions.Result(error=_result)

        return _result

    def test(self, context):
        # TODO(rakhmerov): Implement.
        return None


class MistralHTTPAction(HTTPAction):

    def run(self, context):
        self.headers = self.headers or {}

        exec_ctx = context.execution
        self.headers.update({
            'Mistral-Workflow-Name': exec_ctx.workflow_name,
            'Mistral-Workflow-Execution-Id': exec_ctx.workflow_execution_id,
            'Mistral-Task-Id': exec_ctx.task_execution_id,
            'Mistral-Action-Execution-Id': exec_ctx.action_execution_id,
            'Mistral-Callback-URL': exec_ctx.callback_url,
        })

        return super(MistralHTTPAction, self).run(context)

    def is_sync(self):
        return False

    def test(self, context):
        return None


class SendEmailAction(actions.Action):
    def __init__(self, from_addr, to_addrs, smtp_server, reply_to=None,
                 cc_addrs=None, bcc_addrs=None, smtp_password=None,
                 subject=None, body=None, html_body=None):
        super(SendEmailAction, self).__init__()
        # TODO(dzimine): validate parameters

        # Task invocation parameters.
        self.to = to_addrs
        self.cc = cc_addrs or []
        self.bcc = bcc_addrs or []
        self.reply_to = reply_to or []
        self.subject = subject or "<No subject>"
        self.body = body or "<No body>"
        self.html_body = html_body

        # Action provider settings.
        self.smtp_server = smtp_server
        self.sender = from_addr
        self.password = smtp_password

    def run(self, context):
        LOG.info(
            "Sending email message "
            "[from=%s, to=%s, reply_to=%s, cc=%s, bcc=%s, subject=%s, "
            "using smtp=%s, body=%s...]",
            self.sender,
            self.to,
            self.reply_to,
            self.cc,
            self.bcc,
            self.subject,
            self.smtp_server,
            self.body[:128]
        )
        if not self.html_body:
            message = text.MIMEText(self.body, _charset='utf-8')
        else:
            message = multipart.MIMEMultipart('alternative')
            message.attach(text.MIMEText(self.body,
                                         'plain',
                                         _charset='utf-8'))
            message.attach(text.MIMEText(self.html_body,
                                         'html',
                                         _charset='utf-8'))
        message['Subject'] = header.Header(self.subject, 'utf-8')
        message['From'] = self.sender
        message['Reply-To'] = header.Header(', '.join(self.reply_to))
        message['To'] = ', '.join(self.to)

        if self.cc:
            message['cc'] = ', '.join(self.cc)

        rcpt = self.cc + self.bcc + self.to

        try:
            s = smtplib.SMTP(self.smtp_server)

            if self.password is not None:
                # Sequence to request TLS connection and log in (RFC-2487).
                s.ehlo()
                s.starttls()
                s.ehlo()
                s.login(self.sender, self.password)

            s.sendmail(from_addr=self.sender,
                       to_addrs=rcpt,
                       msg=message.as_string())
        except (smtplib.SMTPException, IOError) as e:
            raise exc.ActionException("Failed to send an email message: %s"
                                      % e)

    def test(self, context):
        # Just logging the operation since this action is not supposed
        # to return a result.
        LOG.info(
            "Sending email message "
            "[from=%s, to=%s, reply_to=%s, cc=%s, bcc=%s, subject=%s, "
            " using smtp=%s, body=%s...]",
            self.sender,
            self.to,
            self.reply_to,
            self.cc,
            self.bcc,
            self.subject,
            self.smtp_server,
            self.body[:128]
        )


class SSHAction(actions.Action):
    """Runs Secure Shell (SSH) command on provided single or multiple hosts.

    It is allowed to provide either a single host or a list of hosts in
    action parameter 'host'. In case of a single host the action result
    will be a single value, otherwise a list of results provided in the
    same order as provided hosts.
    """
    @property
    def _execute_cmd_method(self):
        return ssh_utils.execute_command

    def __init__(self, cmd, host, username,
                 password="", private_key_filename=None, private_key=None):
        super(SSHAction, self).__init__()

        self.cmd = cmd
        self.host = host
        self.username = username
        self.password = password
        self.private_key_filename = private_key_filename
        self.private_key = private_key

        self.params = {
            'cmd': self.cmd,
            'host': self.host,
            'username': self.username,
            'password': self.password,
            'private_key_filename': self.private_key_filename,
            'private_key': self.private_key
        }

    def run(self, context):
        def raise_exc(parent_exc=None):
            message = ("Failed to execute ssh cmd "
                       "'%s' on %s" % (self.cmd, self.host))
            # We suppress the actual parent error messages in favor of
            # more generic ones as we might be leaking information to the CLI
            if parent_exc:
                # The full error message needs to be logged regardless
                LOG.exception(message + " Exception: %s", str(parent_exc))
            raise exc.ActionException(message)

        try:
            results = []

            if not isinstance(self.host, list):
                self.host = [self.host]

            for host_name in self.host:
                self.params['host'] = host_name

                status_code, result = self._execute_cmd_method(**self.params)

                if status_code > 0:
                    return raise_exc()
                else:
                    results.append(result)

            if len(results) > 1:
                return results

            return result
        except Exception as e:
            return raise_exc(parent_exc=e)

    def test(self, context):
        return utils.to_json_str(self.params)


class SSHProxiedAction(SSHAction):
    @property
    def _execute_cmd_method(self):
        return ssh_utils.execute_command_via_gateway

    def __init__(self, cmd, host, username, private_key_filename,
                 gateway_host, gateway_username=None,
                 password=None, proxy_command=None,
                 private_key=None):
        super(SSHProxiedAction, self).__init__(
            cmd,
            host,
            username,
            password,
            private_key_filename,
            private_key
        )

        self.gateway_host = gateway_host
        self.gateway_username = gateway_username

        self.params.update(
            {
                'gateway_host': gateway_host,
                'gateway_username': gateway_username,
                'proxy_command': proxy_command
            }
        )


class JavaScriptAction(actions.Action):
    """Evaluates given JavaScript.

    """

    def __init__(self, script, context=None):
        """Context here refers to a javasctript context

        Not the usual mistral context. That is passed during the run method
        """
        super(JavaScriptAction, self).__init__()

        self.script = script
        self.js_context = context

    def run(self, context):
        try:
            script = """function f() {
                %s
            }
            f()
            """ % self.script
            return javascript.evaluate(script, self.js_context)
        except Exception as e:
            raise exc.ActionException("JavaScriptAction failed: %s" % str(e))

    def test(self, context):
        return self.script


class SleepAction(actions.Action):
    """Sleep action.

    This action sleeps for given amount of seconds. It can be mostly useful
    for testing and debugging purposes.
    """
    def __init__(self, seconds=1):
        super(SleepAction, self).__init__()

        try:
            self._seconds = int(seconds)
            self._seconds = 0 if self._seconds < 0 else self._seconds
        except ValueError:
            self._seconds = 0

    def run(self, context):
        LOG.info('Running sleep action [seconds=%s]', self._seconds)

        time.sleep(self._seconds)

        return None

    def test(self, context):
        time.sleep(1)

        return None


class TestDictAction(actions.Action):
    """Generates test dict."""

    def __init__(self, size=0, key_prefix='', val=''):
        super(TestDictAction, self).__init__()

        self.size = size
        self.key_prefix = key_prefix
        self.val = val

    def run(self, context):
        LOG.info(
            'Running test_dict action [size=%s, key_prefix=%s, val=%s]',
            self.size,
            self.key_prefix,
            self.val
        )

        res = {}

        for i in range(self.size):
            res['%s%s' % (self.key_prefix, i)] = self.val

        return res

    def test(self, context):
        return {}



# Copyright 2016 - Nokia Networks
# Copyright 2017 - Brocade Communications Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_log import log as logging

from mistral import config as cfg
from mistral.event_engine import default_event_engine as evt_eng
from mistral.rpc import base as rpc
from mistral.service import base as service_base
from mistral.utils import profiler as profiler_utils


LOG = logging.getLogger(__name__)


class EventEngineServer(service_base.MistralService):
    """RPC EventEngine server.

    This class manages event engine life-cycle and gets registered as
    an RPC endpoint to process event engine specific calls. It also
    registers a cluster member associated with this instance of event
    engine.
    """

    def __init__(self, event_engine):
        super(EventEngineServer, self).__init__('event-engine_group')

        self._event_engine = event_engine
        self._rpc_server = None

    def start(self):
        super(EventEngineServer, self).start()

        profiler_utils.setup(
            'mistral-event-engine',
            cfg.CONF.event_engine.host
        )

        # Initialize and start RPC server.

        self._rpc_server = rpc.get_rpc_server_driver()(cfg.CONF.event_engine)
        self._rpc_server.register_endpoint(self)

        self._rpc_server.run()

        self._notify_started('Event engine server started.')

    def stop(self, graceful=False):
        super(EventEngineServer, self).stop(graceful)

        if self._rpc_server:
            self._rpc_server.stop(graceful)

    def create_event_trigger(self, rpc_ctx, trigger, events):
        LOG.info(
            "Received RPC request 'create_event_trigger'[rpc_ctx=%s,"
            " trigger=%s, events=%s", rpc_ctx, trigger, events
        )

        return self._event_engine.create_event_trigger(trigger, events)

    def delete_event_trigger(self, rpc_ctx, trigger, events):
        LOG.info(
            "Received RPC request 'delete_event_trigger'[rpc_ctx=%s,"
            " trigger=%s, events=%s", rpc_ctx, trigger, events
        )

        return self._event_engine.delete_event_trigger(trigger, events)

    def update_event_trigger(self, rpc_ctx, trigger):
        LOG.info(
            "Received RPC request 'update_event_trigger'[rpc_ctx=%s,"
            " trigger=%s", rpc_ctx, trigger
        )

        return self._event_engine.update_event_trigger(trigger)


def get_oslo_service():
    return EventEngineServer(evt_eng.DefaultEventEngine())



# Copyright 2013 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import copy
import re

import yaml

from mistral import exceptions as exc
from mistral.lang.v2 import workbook
from mistral.tests.unit.lang.v2 import base


class WorkbookSpecValidationTest(base.WorkbookSpecValidationTestCase):

    def test_build_valid_workbook_spec(self):
        wb_spec = self._parse_dsl_spec(dsl_file='my_workbook.yaml')

        # Workbook.
        act_specs = wb_spec.get_actions()
        wf_specs = wb_spec.get_workflows()

        self.assertEqual('2.0', wb_spec.get_version())
        self.assertEqual('my_workbook', wb_spec.get_name())
        self.assertEqual('This is a test workbook', wb_spec.get_description())
        self.assertListEqual(['test', 'v2'], wb_spec.get_tags())
        self.assertIsNotNone(act_specs)
        self.assertIsNotNone(wf_specs)

        # Actions.
        action_spec = act_specs.get('action1')

        self.assertIsNotNone(action_spec)
        self.assertEqual('2.0', action_spec.get_version())
        self.assertEqual('action1', action_spec.get_name())
        self.assertEqual(
            'This is a test ad-hoc action',
            action_spec.get_description()
        )
        self.assertListEqual(['test', 'v2'], action_spec.get_tags())
        self.assertEqual('std.echo', action_spec.get_base())
        self.assertDictEqual(
            {'output': 'Hello <% $.name %>!'},
            action_spec.get_base_input()
        )
        self.assertDictEqual({}, action_spec.get_input())
        self.assertEqual('<% $ %>', action_spec.get_output())

        # Workflows.

        self.assertEqual(2, len(wf_specs))

        wf1_spec = wf_specs.get('wf1')

        self.assertEqual('2.0', wf1_spec.get_version())
        self.assertEqual('wf1', wf1_spec.get_name())
        self.assertEqual(
            'This is a test workflow',
            wf1_spec.get_description()
        )
        self.assertListEqual(['test', 'v2'], wf1_spec.get_tags())
        self.assertEqual('reverse', wf1_spec.get_type())
        self.assertEqual(2, len(wf1_spec.get_tasks()))

        # Tasks.

        task1_spec = wf1_spec.get_tasks().get('task1')

        self.assertIsNotNone(task1_spec)
        self.assertEqual('2.0', task1_spec.get_version())
        self.assertEqual('task1', task1_spec.get_name())
        self.assertEqual('This is a test task', task1_spec.get_description())
        self.assertEqual('action1', task1_spec.get_action_name())
        self.assertEqual({'name': '<% $.name %>'}, task1_spec.get_input())

        policies = task1_spec.get_policies()

        self.assertEqual(2, policies.get_wait_before())
        self.assertEqual(5, policies.get_wait_after())
        self.assertEqual(3, policies.get_concurrency())

        retry_spec = policies.get_retry()

        self.assertEqual(10, retry_spec.get_count())
        self.assertEqual(30, retry_spec.get_delay())
        self.assertEqual('<% $.my_val = 10 %>', retry_spec.get_break_on())

        task2_spec = wf1_spec.get_tasks().get('task2')

        self.assertIsNotNone(task2_spec)
        self.assertEqual('2.0', task2_spec.get_version())
        self.assertEqual('task2', task2_spec.get_name())
        self.assertEqual('std.echo', task2_spec.get_action_name())
        self.assertIsNone(task2_spec.get_workflow_name())
        self.assertEqual(
            {'output': 'Thanks <% $.name %>!'},
            task2_spec.get_input()
        )

        wf2_spec = wf_specs.get('wf2')

        self.assertEqual('2.0', wf2_spec.get_version())
        self.assertEqual('wf2', wf2_spec.get_name())
        self.assertListEqual(['test', 'v2'], wf2_spec.get_tags())
        self.assertEqual('direct', wf2_spec.get_type())
        self.assertEqual(11, len(wf2_spec.get_tasks()))

        task_defaults_spec = wf2_spec.get_task_defaults()

        self.assertListEqual(
            [('fail', '<% $.my_val = 0 %>', {})],
            task_defaults_spec.get_on_error().get_next()
        )
        self.assertListEqual(
            [('pause', '', {})],
            task_defaults_spec.get_on_success().get_next()
        )
        self.assertListEqual(
            [('succeed', '', {})],
            task_defaults_spec.get_on_complete().get_next()
        )

        task3_spec = wf2_spec.get_tasks().get('task3')

        self.assertIsNotNone(task3_spec)
        self.assertEqual('2.0', task3_spec.get_version())
        self.assertEqual('task3', task3_spec.get_name())
        self.assertIsNone(task3_spec.get_action_name())
        self.assertEqual('wf1', task3_spec.get_workflow_name())
        self.assertEqual(
            {
                'name': 'John Doe',
                'age': 32,
                'param1': None,
                'param2': False
            },
            task3_spec.get_input()
        )
        self.assertListEqual(
            [('task4', '<% $.my_val = 1 %>', {})],
            task3_spec.get_on_error().get_next()
        )
        self.assertListEqual(
            [('task5', '<% $.my_val = 2 %>', {})],
            task3_spec.get_on_success().get_next()
        )
        self.assertListEqual(
            [('task6', '<% $.my_val = 3 %>', {})],
            task3_spec.get_on_complete().get_next()
        )

        task7_spec = wf2_spec.get_tasks().get('task7')

        self.assertEqual(
            {
                'is_true': True,
                'object_list': [1, None, 'str'],
                'is_string': '50'
            },
            task7_spec.get_input()
        )

        self.assertEqual(
            {'vm_info': '<% $.vms %>'},
            task7_spec.get_with_items()
        )

        task8_spec = wf2_spec.get_tasks().get('task8')

        self.assertEqual(
            {
                'itemX': '<% $.arrayI %>',
                "itemY": '<% $.arrayJ %>'
            },
            task8_spec.get_with_items()
        )

        self.assertEqual(
            {
                'expr_list': ['<% $.v %>', '<% $.k %>'],
                'expr': '<% $.value %>',
            },
            task8_spec.get_input()
        )

        self.assertEqual('nova', task8_spec.get_target())

        task9_spec = wf2_spec.get_tasks().get('task9')

        self.assertEqual('all', task9_spec.get_join())

        task10_spec = wf2_spec.get_tasks().get('task10')

        self.assertEqual(2, task10_spec.get_join())

        task11_spec = wf2_spec.get_tasks().get('task11')

        self.assertEqual('one', task11_spec.get_join())

        task12_spec = wf2_spec.get_tasks().get('task12')

        self.assertDictEqual(
            {
                'url': 'http://site.com?q=<% $.query %>',
                'params': ''
            },
            task12_spec.get_input()
        )

        task13_spec = wf2_spec.get_tasks().get('task13')

        self.assertEqual('std.noop', task13_spec.get_action_name())
        self.assertEqual('No-op task', task13_spec.get_description())

    def test_adhoc_action_with_base_in_one_string(self):
        wb_spec = self._parse_dsl_spec(dsl_file='my_workbook.yaml')

        act_specs = wb_spec.get_actions()
        action_spec = act_specs.get("action2")

        self.assertEqual('std.echo', action_spec.get_base())
        self.assertEqual(
            {'output': 'Echo output'},
            action_spec.get_base_input()
        )

    def test_spec_to_dict(self):
        wb_spec = self._parse_dsl_spec(dsl_file='my_workbook.yaml')

        d = wb_spec.to_dict()

        self.assertEqual('2.0', d['version'])
        self.assertEqual('2.0', d['workflows']['version'])
        self.assertEqual('2.0', d['workflows']['wf1']['version'])

    def test_version_required(self):
        dsl_dict = copy.deepcopy(self._dsl_blank)
        dsl_dict.pop('version', None)

        # TODO(m4dcoder): Check required property error when v1 is deprecated.
        # The version property is not required for v1 workbook whereas it is
        # a required property in v2. For backward compatibility, if no version
        # is not provided, the workbook spec parser defaults to v1 and the
        # required property exception is not triggered. However, a different
        # spec validation error returns due to drastically different schema
        # between workbook versions.
        self.assertRaises(
            exc.DSLParsingException,
            self._spec_parser,
            yaml.safe_dump(dsl_dict)
        )

    def test_version(self):
        tests = [
            ({'version': None}, True),
            ({'version': ''}, True),
            ({'version': '1.0'}, True),
            ({'version': '2.0'}, False),
            ({'version': 2.0}, False),
            ({'version': 2}, False)
        ]

        for version, expect_error in tests:
            self._parse_dsl_spec(changes=version, expect_error=expect_error)

    def test_name_required(self):
        dsl_dict = copy.deepcopy(self._dsl_blank)
        dsl_dict.pop('name', None)

        exception = self.assertRaises(
            exc.DSLParsingException,
            self._spec_parser,
            yaml.safe_dump(dsl_dict)
        )

        self.assertIn("'name' is a required property", str(exception))

    def test_name(self):
        tests = [
            ({'name': ''}, True),
            ({'name': None}, True),
            ({'name': 12345}, True),
            ({'name': 'foobar'}, False)
        ]

        for name, expect_error in tests:
            self._parse_dsl_spec(changes=name, expect_error=expect_error)

    def test_description(self):
        tests = [
            ({'description': ''}, True),
            ({'description': None}, True),
            ({'description': 12345}, True),
            ({'description': 'This is a test workflow.'}, False)
        ]

        for description, expect_error in tests:
            self._parse_dsl_spec(
                changes=description,
                expect_error=expect_error
            )

    def test_tags(self):
        tests = [
            ({'tags': ''}, True),
            ({'tags': ['']}, True),
            ({'tags': None}, True),
            ({'tags': 12345}, True),
            ({'tags': ['foo', 'bar']}, False),
            ({'tags': ['foobar', 'foobar']}, True)
        ]

        for tags, expect_error in tests:
            self._parse_dsl_spec(changes=tags, expect_error=expect_error)

    def test_actions(self):
        actions = {
            'version': '2.0',
            'noop': {
                'base': 'std.noop'
            },
            'echo': {
                'base': 'std.echo'
            }
        }

        tests = [
            ({'actions': []}, True),
            ({'actions': {}}, True),
            ({'actions': None}, True),
            ({'actions': {'version': None}}, True),
            ({'actions': {'version': ''}}, True),
            ({'actions': {'version': '1.0'}}, True),
            ({'actions': {'version': '2.0'}}, False),
            ({'actions': {'version': 2.0}}, False),
            ({'actions': {'version': 2}}, False),
            ({'actions': {'noop': actions['noop']}}, False),
            ({'actions': {'version': '2.0', 'noop': 'std.noop'}}, True),
            ({'actions': actions}, False)
        ]

        for adhoc_actions, expect_error in tests:
            self._parse_dsl_spec(
                changes=adhoc_actions,
                expect_error=expect_error
            )

    def test_workflows(self):
        workflows = {
            'version': '2.0',
            'wf1': {
                'tasks': {
                    'noop': {
                        'action': 'std.noop'
                    }
                }
            },
            'wf2': {
                'tasks': {
                    'echo': {
                        'action': 'std.echo output="This is a test."'
                    }
                }
            }
        }

        tests = [
            # ({'workflows': []}, True),
            # ({'workflows': {}}, True),
            # ({'workflows': None}, True),
            # ({'workflows': {'version': None}}, True),
            # ({'workflows': {'version': ''}}, True),
            # ({'workflows': {'version': '1.0'}}, True),
            # ({'workflows': {'version': '2.0'}}, False),
            # ({'workflows': {'version': 2.0}}, False),
            # ({'workflows': {'version': 2}}, False),
            # ({'workflows': {'wf1': workflows['wf1']}}, False),
            ({'workflows': {'version': '2.0', 'wf1': 'wf1'}}, True),
            ({'workflows': workflows}, False)
        ]

        for workflows, expect_error in tests:
            self._parse_dsl_spec(changes=workflows, expect_error=expect_error)

    def test_workflow_name_validation(self):
        wb_spec = self._parse_dsl_spec(dsl_file='workbook_schema_test.yaml')

        d = wb_spec.to_dict()

        self.assertEqual('2.0', d['version'])
        self.assertEqual('2.0', d['workflows']['version'])

        workflow_names = ['workflowversion', 'versionworkflow',
                          'workflowversionworkflow', 'version_workflow']

        action_names = ['actionversion', 'versionaction',
                        'actionversionaction']

        for name in workflow_names:
            self.assertEqual('2.0', d['workflows'][name]['version'])
            self.assertEqual(name, d['workflows'][name]['name'])

        for name in action_names:
            self.assertEqual('2.0', d['actions'][name]['version'])
            self.assertEqual(name, d['actions'][name]['name'])

    def test_name_regex(self):
        # We want to match a string containing version at any point.
        valid_names = (
            "workflowversion",
            "versionworkflow",
            "workflowversionworkflow",
            "version_workflow",
            "version-workflow",
        )

        for valid in valid_names:
            result = re.match(workbook.NON_VERSION_WORD_REGEX, valid)
            self.assertIsNotNone(
                result,
                "Expected match for: {}".format(valid)
            )

        # ... except, we don't want to match a string that isn't just one word
        # or is exactly "version"
        invalid_names = ("version", "my workflow")

        for invalid in invalid_names:
            result = re.match(workbook.NON_VERSION_WORD_REGEX, invalid)
            self.assertIsNone(
                result,
                "Didn't expected match for: {}".format(invalid)
            )



# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_log import log as logging
import pecan
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from mistral.api.controllers import info
from mistral.api.controllers import maintenance
from mistral.api.controllers import resource
from mistral.api.controllers.v2 import root as v2_root

LOG = logging.getLogger(__name__)

API_STATUS = wtypes.Enum(str, 'SUPPORTED', 'CURRENT', 'DEPRECATED')


class APIVersion(resource.Resource):
    """An API Version."""

    id = wtypes.text
    "The version identifier."

    status = API_STATUS
    "The status of the API (SUPPORTED, CURRENT or DEPRECATED)."

    links = wtypes.ArrayType(resource.Link)
    "The link to the versioned API."

    @classmethod
    def sample(cls):
        return cls(
            id='v1.0',
            status='CURRENT',
            links=[
                resource.Link(target_name='v1', rel="self",
                              href='http://example.com:9777/v1')
            ]
        )


class APIVersions(resource.Resource):
    """API Versions."""
    versions = wtypes.ArrayType(APIVersion)

    @classmethod
    def sample(cls):
        v2 = APIVersion(id='v2.0', status='CURRENT', rel="self",
                        href='http://example.com:9777/v2')
        return cls(versions=[v2])


class RootController(object):
    v2 = v2_root.Controller()
    info = info.InfoController()
    maintenance = maintenance.MaintenanceController()

    @wsme_pecan.wsexpose(APIVersions)
    def index(self):
        LOG.debug("Fetching API versions.")

        host_url_v2 = '%s/%s' % (pecan.request.application_url, 'v2')
        api_v2 = APIVersion(
            id='v2.0',
            status='CURRENT',
            links=[resource.Link(href=host_url_v2, target='v2',
                                 rel="self",)]
        )

        return APIVersions(versions=[api_v2])



# Copyright 2020 - Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import json
from pathlib import Path

from rally.common import cfg
from rally.task import validation
from rally_openstack import consts
from rally_openstack import scenario
from rally_openstack.scenarios.mistral import utils

CONF = cfg.CONF

SCENARIO_TIMEOUT_SEC = 16000

home_dir = str(Path.home())
wf_dir = '%s/.rally/extra/scenarios/big_wf/' % home_dir

action_files = ['dummy_actions.yaml', 'dummy_actions_nuage.yaml']
common_workflow_files = ['sub_wfs.yaml']


class MistralHugeWorkflowScenario(utils.MistralScenario):
    main_wf_file_name = ''
    params_filename = ''
    wf_name = ''
    created_wfs = []
    created_wbs = []

    def run(self):
        namespace = ''
        CONF.openstack.mistral_execution_timeout = SCENARIO_TIMEOUT_SEC

        self.create_common_workflows()
        self.create_actions()
        self.create_main_workflow(namespace=namespace)
        params = self._read_params_from_file()

        self.run_workflow(params, namespace=namespace)
        self.clean_after_scenario(namespace=namespace)

    def clean_after_scenario(self, namespace=''):
        for wf in self.created_wfs:
            self._delete_workflow(wf, namespace=namespace)

        for wb in self.created_wbs:
            self._delete_workbook(wb, namespace=namespace)

    def create_common_workflows(self):
        for file in common_workflow_files:
            with open(wf_dir + file, 'r+') as f:
                definition = f.read()
                wfs = self._create_workflow(definition)

                for wf in wfs:
                    self.created_wfs.append(wf.name)

    def create_actions(self):
        for file in action_files:
            with open(wf_dir + file, 'r+') as f:
                definition = f.read()
                wb = self._create_workbook(definition)

                self.created_wbs.append(wb.name)

    def _create_workbook(self, definition, namespace=''):
        return self.clients("mistral").workbooks.create(
            definition,
            namespace=namespace
        )

    def _read_params_from_file(self):
        with open(wf_dir + self.params_filename, 'r+') as f:
            params_string = f.read()
            params = json.loads(params_string)

            return params

    def run_workflow(self, params, namespace=''):
        input = {}

        self._create_execution(
            self.wf_name,
            wf_input=input,
            namespace=namespace,
            **params
        )

    def create_main_workflow(self, namespace=''):
        with open(wf_dir + self.main_wf_file_name, 'r+') as f:
            definition = f.read()
            wfs = self._create_workflow(definition, namespace=namespace)

            for wf in wfs:
                self.created_wfs.append(wf.name)


@validation.add("required_platform", platform="openstack", users=True)
@validation.add("required_services", services=[consts.Service.MISTRAL])
@scenario.configure(name="MistralExecutions.TerminateScenario",
                    platform="openstack")
class TerminateScenario(MistralHugeWorkflowScenario):
    main_wf_file_name = 'terminate_wf.yaml'
    params_filename = 'terminate_params.json'
    wf_name = 'mistral_cmg_terminate'


@validation.add("required_platform", platform="openstack", users=True)
@validation.add("required_services", services=[consts.Service.MISTRAL])
@scenario.configure(name="MistralExecutions.DeployScenario",
                    platform="openstack")
class DeployScenario(MistralHugeWorkflowScenario):
    main_wf_file_name = 'deploy_wf.yaml'
    params_filename = 'deploy_params.json'
    wf_name = 'mistral_cmg_deploy'



# Copyright 2020 Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock

from mistral_lib import actions as ml_actions
from mistral_lib.utils import inspect_utils

from mistral.actions import legacy
from mistral.tests.unit import base


class BuildMessageAction(ml_actions.Action):
    msg_pattern = '%s'

    def __init__(self, name):
        super(BuildMessageAction, self).__init__()

        self._name = name

    def run(self, context):
        return self.msg_pattern % self._name


class TestActionGenerator(object):
    base_action_class = BuildMessageAction

    @classmethod
    def create_actions(cls):
        action_dicts = []

        hello_action_cls = type(
            'HelloAction',
            (BuildMessageAction,),
            {'msg_pattern': 'Hello, %s!'}
        )

        action_dicts.append(
            {
                'class': hello_action_cls,
                'name': 'hello',
                'description': 'The action builds a hello message',
                'arg_list': inspect_utils.get_arg_list_as_str(
                    hello_action_cls.__init__
                )
            }
        )

        goodbye_action_cls = type(
            'GoodbyeAction',
            (BuildMessageAction,),
            {'msg_pattern': 'Goodbye, %s!'}
        )

        action_dicts.append(
            {
                'class': goodbye_action_cls,
                'name': 'goodbye',
                'description': 'The action builds a goodbye message',
                'arg_list': inspect_utils.get_arg_list_as_str(
                    goodbye_action_cls.__init__
                )
            }
        )

        return action_dicts


class LegacyActionProviderTest(base.BaseTest):
    def test_only_builtin_actions(self):
        self.override_config(
            'load_action_generators',
            False,
            'legacy_action_provider'
        )
        self.override_config(
            'only_builtin_actions',
            True,
            'legacy_action_provider'
        )

        provider = legacy.LegacyActionProvider()

        # Test find_all() method.
        action_descs = provider.find_all()

        self.assertTrue(len(action_descs) > 0)

        self.assertTrue(
            all(
                [
                    a_d.action_class.__module__.startswith('mistral.')
                    for a_d in action_descs
                ]
            )
        )

        self._assert_single_item(action_descs, name='std.echo')

        # Test find() method.
        action_desc = provider.find('std.echo')

        self.assertIsNotNone(action_desc)
        self.assertEqual('std.echo', action_desc.name)
        self.assertIn('Echo action.', action_desc.description)
        self.assertEqual(
            'mistral.actions.std_actions.EchoAction',
            action_desc.action_class_name
        )
        self.assertEqual('output, delay=0', action_desc.params_spec)

    @mock.patch.object(
        legacy.LegacyActionProvider,
        '_get_action_generators',
        mock.MagicMock(return_value=[TestActionGenerator])
    )
    def test_only_action_plugins(self):
        self.override_config(
            'load_action_generators',
            False,
            'legacy_action_provider'
        )

        provider = legacy.LegacyActionProvider()

        action_descs = provider.find_all()

        prefix = 'mistral.actions.std_actions'

        self.assertTrue(
            all(
                [
                    a_d.action_class.__module__ == prefix
                    for a_d in action_descs
                ]
            )
        )

        self._assert_single_item(action_descs, name='std.echo')

    @mock.patch.object(
        legacy.LegacyActionProvider,
        '_get_action_generators',
        mock.MagicMock(return_value=[TestActionGenerator])
    )
    def test_only_action_generators(self):
        self.override_config(
            'load_action_generators',
            True,
            'legacy_action_provider'
        )
        self.override_config(
            'load_action_plugins',
            False,
            'legacy_action_provider'
        )

        provider = legacy.LegacyActionProvider()

        action_descs = provider.find_all()

        self.assertEqual(2, len(action_descs))

        hello_action_desc = self._assert_single_item(
            action_descs,
            name='hello',
            params_spec='name',
            description='The action builds a hello message',
            action_class_name='mistral.tests.unit.actions.'
                              'test_legacy_action_provider.BuildMessageAction',
            action_class_attributes={'msg_pattern': 'Hello, %s!'}
        )

        hello_action = hello_action_desc.instantiate({'name': 'Forest'}, {})

        self.assertEqual('Hello, Forest!', hello_action.run(None))

        goodbye_action_desc = self._assert_single_item(
            action_descs,
            name='goodbye',
            params_spec='name',
            description='The action builds a goodbye message',
            action_class_name='mistral.tests.unit.actions.'
                              'test_legacy_action_provider.BuildMessageAction',
            action_class_attributes={'msg_pattern': 'Goodbye, %s!'}
        )

        goodbye_action = goodbye_action_desc.instantiate(
            {'name': 'Lieutenant Dan'},
            {}
        )

        self.assertEqual('Goodbye, Lieutenant Dan!', goodbye_action.run(None))



#  Copyright 2019 - Nokia Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from unittest import TestCase

from mistral.utils import safe_yaml


class TestSafeLoader(TestCase):
    def test_safe_load(self):
        yaml_text = """
        version: '2.0'

        wf1:
          type: direct

          input:
            - a: &a ["lol","lol","lol","lol","lol"]
            - b: &b [*a,*a,*a,*a,*a,*a,*a,*a,*a]
            - c: &c [*b,*b,*b,*b,*b,*b,*b,*b,*b]
            - d: &d [*c,*c,*c,*c,*c,*c,*c,*c,*c]
            - e: &e [*d,*d,*d,*d,*d,*d,*d,*d,*d]
            - f: &f [*e,*e,*e,*e,*e,*e,*e,*e,*e]
            - g: &g [*f,*f,*f,*f,*f,*f,*f,*f,*f]
            - h: &h [*g,*g,*g,*g,*g,*g,*g,*g,*g]
            - i: &i [*h,*h,*h,*h,*h,*h,*h,*h,*h]


          tasks:
            hello:
              action: std.echo output="Hello"
              wait-before: 1
              publish:
                result: <% task(hello).result %>
        """

        result = {
            'version': '2.0',
            'wf1':
                {'type': 'direct',
                 'input': [
                     {'a': '&a ["lol","lol","lol","lol","lol"]'},
                     {'b': '&b [*a,*a,*a,*a,*a,*a,*a,*a,*a]'},
                     {'c': '&c [*b,*b,*b,*b,*b,*b,*b,*b,*b]'},
                     {'d': '&d [*c,*c,*c,*c,*c,*c,*c,*c,*c]'},
                     {'e': '&e [*d,*d,*d,*d,*d,*d,*d,*d,*d]'},
                     {'f': '&f [*e,*e,*e,*e,*e,*e,*e,*e,*e]'},
                     {'g': '&g [*f,*f,*f,*f,*f,*f,*f,*f,*f]'},
                     {'h': '&h [*g,*g,*g,*g,*g,*g,*g,*g,*g]'},
                     {'i': '&i [*h,*h,*h,*h,*h,*h,*h,*h,*h]'}],
                 'tasks':
                    {'hello': {
                        'action': 'std.echo output="Hello"',
                        'wait-before': 1, 'publish':
                            {'result': '<% task(hello).result %>'}
                    }}
                 }
        }
        self.assertEqual(result, safe_yaml.load(yaml_text))



# Copyright 2015 - Mirantis, Inc.
# Copyright 2017 - Brocade Communications Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import abc
from functools import wraps

from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_messaging.exceptions import MessagingTimeout
from oslo_messaging.rpc import client
from stevedore import driver

from mistral import exceptions as exc


LOG = logging.getLogger(__name__)


_IMPL_CLIENT = None
_IMPL_SERVER = None
_TRANSPORT = None


def cleanup():
    """Intended to be used by tests to recreate all RPC related objects."""

    global _TRANSPORT

    _TRANSPORT = None


# TODO(rakhmerov): This method seems misplaced. Now we have different kind
# of transports (oslo, kombu) and this module should not have any oslo
# specific things anymore.
def get_transport():
    global _TRANSPORT

    if not _TRANSPORT:
        _TRANSPORT = messaging.get_rpc_transport(cfg.CONF)

    return _TRANSPORT


def get_rpc_server_driver():
    rpc_impl = cfg.CONF.rpc_implementation

    global _IMPL_SERVER
    if not _IMPL_SERVER:
        _IMPL_SERVER = driver.DriverManager(
            'mistral.rpc.backends',
            '%s_server' % rpc_impl
        ).driver

    return _IMPL_SERVER


def get_rpc_client_driver():
    rpc_impl = cfg.CONF.rpc_implementation

    global _IMPL_CLIENT
    if not _IMPL_CLIENT:
        _IMPL_CLIENT = driver.DriverManager(
            'mistral.rpc.backends',
            '%s_client' % rpc_impl
        ).driver

    return _IMPL_CLIENT


def _wrap_exception_and_reraise(exception):
    message = "%s: %s" % (exception.__class__.__name__, exception.args[0])

    raise exc.MistralException(message)


def wrap_messaging_exception(method):
    """The decorator unwraps a remote error into one of the mistral exceptions.

    oslo.messaging has different behavior on raising exceptions depending on
    whether we use 'fake' or 'rabbit' transports. In case of 'rabbit' transport
    it raises an instance of RemoteError which forwards directly to the API.
    The RemoteError instance contains one of the MistralException instances
    raised remotely on the RPC server side and for correct exception handling
    we need to unwrap and raise the original wrapped exception.
    """

    @wraps(method)
    def decorator(*args, **kwargs):
        try:
            return method(*args, **kwargs)
        except exc.MistralException:
            raise
        except MessagingTimeout:
            timeout = cfg.CONF.rpc_response_timeout
            raise exc.MistralException('This rpc call "%s" took longer than '
                                       'configured %s seconds.' %
                                       (method.__name__, timeout))
        except (client.RemoteError, exc.KombuException, Exception) as e:
            # Since we're going to transform the original exception
            # we need to log it as is.
            LOG.exception(
                "Caught a messaging remote error."
                " See details of the original exception."
            )

            if hasattr(e, 'exc_type') and hasattr(exc, e.exc_type):
                exc_cls = getattr(exc, e.exc_type)

                raise exc_cls(e.value)

            _wrap_exception_and_reraise(e)

    return decorator


class RPCClient(object):
    def __init__(self, conf):
        """Base class for RPCClient's drivers

        RPC Client is responsible for sending requests to RPC Server.
        All RPC client drivers have to inherit from this class.

        :param conf: Additional config provided by upper layer.
        """
        self.conf = conf

    @abc.abstractmethod
    def sync_call(self, ctx, method, target=None, **kwargs):
        """Synchronous call of RPC method.

        Blocks the thread and wait for method result.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def async_call(self, ctx, method, target=None, fanout=False, **kwargs):
        """Asynchronous call of RPC method.

        Does not block the thread, just send invoking data to
        the RPC server and immediately returns nothing.
        """
        raise NotImplementedError


class RPCServer(object):
    def __init__(self, conf):
        """Base class for RPCServer's drivers

        RPC Server should listen for request coming from RPC Clients and
        respond to them respectively to the registered endpoints.
        All RPC server drivers have to inherit from this class.

        :param conf: Additional config provided by upper layer.
        """
        self.conf = conf

    @abc.abstractmethod
    def register_endpoint(self, endpoint):
        """Registers a new RPC endpoint.

        :param endpoint: an object containing methods which
         will be used as RPC methods.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def run(self, executor='eventlet'):
        """Runs the RPC server.

        :param executor: Executor used to process incoming requests. Different
            implementations may support different options.
        """
        raise NotImplementedError

    def stop(self, graceful=False):
        """Stop the RPC server.

        :param graceful: True if this method call should wait till all
            internal threads are finished.
        :return:
        """
        # No-op by default.
        pass

    def wait(self):
        """Wait till all internal threads are finished."""
        # No-op by default.
        pass



# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral.db import utils as db_utils
from mistral.db.v2 import api as db_api
from mistral.engine import base
from mistral.engine import post_tx_queue
from mistral.engine import workflow_handler as wf_handler
from mistral import expressions
from mistral.scheduler import base as sched_base
from mistral.utils import wf_trace
from mistral.workflow import data_flow
from mistral.workflow import states


_CONTINUE_TASK_PATH = 'mistral.engine.policies._continue_task'
_COMPLETE_TASK_PATH = 'mistral.engine.policies._complete_task'
_FAIL_IF_INCOMPLETE_TASK_PATH = (
    'mistral.engine.policies._fail_task_if_incomplete'
)


def build_policies(policies_spec, wf_spec):
    task_defaults = wf_spec.get_task_defaults()
    wf_policies = task_defaults.get_policies() if task_defaults else None

    if not (policies_spec or wf_policies):
        return []

    return construct_policies_list(policies_spec, wf_policies)


def get_policy_factories():
    return [
        build_pause_before_policy,
        build_wait_before_policy,
        build_wait_after_policy,
        build_fail_on_policy,
        build_retry_policy,
        build_timeout_policy,
        build_concurrency_policy
    ]


def construct_policies_list(policies_spec, wf_policies):
    policies = []

    for factory in get_policy_factories():
        policy = factory(policies_spec)

        if wf_policies and not policy:
            policy = factory(wf_policies)

        if policy:
            policies.append(policy)

    return policies


def build_wait_before_policy(policies_spec):
    if not policies_spec:
        return None

    wait_before = policies_spec.get_wait_before()

    if isinstance(wait_before, str) or wait_before > 0:
        return WaitBeforePolicy(wait_before)
    else:
        return None


def build_wait_after_policy(policies_spec):
    if not policies_spec:
        return None

    wait_after = policies_spec.get_wait_after()

    if isinstance(wait_after, str) or wait_after > 0:
        return WaitAfterPolicy(wait_after)
    else:
        return None


def build_retry_policy(policies_spec):
    if not policies_spec:
        return None

    retry = policies_spec.get_retry()

    if not retry:
        return None

    return RetryPolicy(
        retry.get_count(),
        retry.get_delay(),
        retry.get_break_on(),
        retry.get_continue_on()
    )


def build_timeout_policy(policies_spec):
    if not policies_spec:
        return None

    timeout_policy = policies_spec.get_timeout()

    if isinstance(timeout_policy, str) or timeout_policy > 0:
        return TimeoutPolicy(timeout_policy)
    else:
        return None


def build_pause_before_policy(policies_spec):
    if not policies_spec:
        return None

    pause_before_policy = policies_spec.get_pause_before()

    return (PauseBeforePolicy(pause_before_policy)
            if pause_before_policy else None)


def build_concurrency_policy(policies_spec):
    if not policies_spec:
        return None

    concurrency_policy = policies_spec.get_concurrency()

    return (ConcurrencyPolicy(concurrency_policy)
            if concurrency_policy else None)


def build_fail_on_policy(policies_spec):
    if not policies_spec:
        return None

    fail_on_policy = policies_spec.get_fail_on()

    return (FailOnPolicy(fail_on_policy)
            if fail_on_policy else None)


def _ensure_context_has_key(runtime_context, key):
    if not runtime_context:
        runtime_context = {}

    if key not in runtime_context:
        runtime_context.update({key: {}})

    return runtime_context


class WaitBeforePolicy(base.TaskPolicy):
    _schema = {
        "properties": {
            "delay": {
                "type": "integer",
                "minimum": 0
            }
        }
    }

    def __init__(self, delay):
        self.delay = delay

    def before_task_start(self, task):
        super(WaitBeforePolicy, self).before_task_start(task)

        # No need to wait for a task if delay is 0
        if self.delay == 0:
            return

        ctx_key = 'wait_before_policy'

        policy_ctx = task.get_policy_context(ctx_key)

        if policy_ctx.get('skip'):
            # Unset state 'RUNNING_DELAYED'.
            task.set_state(states.RUNNING, None)

            return

        if task.get_state() != states.IDLE:
            policy_ctx.update({'skip': True})

            task.set_state(
                states.RUNNING_DELAYED,
                "Delayed by 'wait-before' policy [delay=%s]" % self.delay
            )

            sched = sched_base.get_system_scheduler()

            job = sched_base.SchedulerJob(
                run_after=self.delay,
                func_name=_CONTINUE_TASK_PATH,
                func_args={'task_ex_id': task.get_id()}
            )

            sched.schedule(job)


class WaitAfterPolicy(base.TaskPolicy):
    _schema = {
        "properties": {
            "delay": {
                "type": "integer",
                "minimum": 0
            }
        }
    }

    def __init__(self, delay):
        self.delay = delay

    def after_task_complete(self, task):
        super(WaitAfterPolicy, self).after_task_complete(task)

        # No need to postpone a task if delay is 0
        if self.delay == 0:
            return

        ctx_key = 'wait_after_policy'

        policy_ctx = task.get_policy_context(ctx_key)

        if policy_ctx.get('skip'):
            # Skip, already processed.
            return

        policy_ctx.update({'skip': True})

        end_state = task.get_state()
        end_state_info = task.get_state_info()

        # Set task state to 'RUNNING_DELAYED'.
        task.set_state(
            states.RUNNING_DELAYED,
            "Delayed by 'wait-after' policy [delay=%s]" % self.delay
        )

        # Schedule to change task state to RUNNING again.
        sched = sched_base.get_system_scheduler()

        job = sched_base.SchedulerJob(
            run_after=self.delay,
            func_name=_COMPLETE_TASK_PATH,
            func_args={
                'task_ex_id': task.get_id(),
                'state': end_state,
                'state_info': end_state_info
            }
        )

        sched.schedule(job)


class RetryPolicy(base.TaskPolicy):
    _schema = {
        "properties": {
            "delay": {
                "type": "integer",
                "minimum": 0
            },
            "count": {
                "type": "integer",
                "minimum": 0
            },
        }
    }

    def __init__(self, count, delay, break_on, continue_on):
        self.count = count
        self.delay = delay
        self._break_on_clause = break_on
        self._continue_on_clause = continue_on

    def after_task_complete(self, task):
        """Possible Cases:

        1. state = SUCCESS
           if continue_on is not specified,
           no need to move to next iteration;
           if current:count achieve retry:count then policy
           breaks the loop (regardless on continue-on condition);
           otherwise - check continue_on condition and if
           it is True - schedule the next iteration,
           otherwise policy breaks the loop.
        2. retry:count = 5, current:count = 2, state = ERROR,
           state = IDLE/DELAYED, current:count = 3
        3. retry:count = 5, current:count = 4, state = ERROR
        Iterations complete therefore state = #{state}, current:count = 4.
        """
        super(RetryPolicy, self).after_task_complete(task)

        # There is nothing to repeat
        if self.count == 0:
            return

        # TODO(m4dcoder): If the task_ex.action_executions and
        # task_ex.workflow_executions collection are not called,
        # then the retry_no in the runtime_context of the task_ex will not
        # be updated accurately. To be exact, the retry_no will be one
        # iteration behind.
        ex = task.task_ex.executions  # noqa

        ctx_key = 'retry_task_policy'

        expr_ctx = task.get_expression_context(
            ctx=data_flow.evaluate_task_outbound_context(task.task_ex)
        )

        continue_on_evaluation = expressions.evaluate(
            self._continue_on_clause,
            expr_ctx
        )

        break_on_evaluation = expressions.evaluate(
            self._break_on_clause,
            expr_ctx
        )

        state = task.get_state()

        if not states.is_completed(state) or states.is_cancelled(state):
            return

        policy_ctx = task.get_policy_context(ctx_key)

        retry_no = 0

        if 'retry_no' in policy_ctx:
            retry_no = policy_ctx['retry_no']

            del policy_ctx['retry_no']

        retries_remain = retry_no < self.count

        stop_continue_flag = (
            task.get_state() == states.SUCCESS and
            not self._continue_on_clause
        )

        stop_continue_flag = (
            stop_continue_flag or
            (self._continue_on_clause and not continue_on_evaluation)
        )

        break_triggered = (
            task.get_state() == states.ERROR and
            break_on_evaluation
        )

        if not retries_remain or break_triggered or stop_continue_flag:
            return

        task.invalidate_result()

        policy_ctx['retry_no'] = retry_no + 1

        task.touch_runtime_context()

        # NOTE(vgvoleg): join tasks in direct workflows can't be
        # retried as-is, because these tasks can't start without
        # a correct logical state.
        if hasattr(task.task_spec, "get_join") and task.task_spec.get_join():
            # TODO(rakhmerov): This is an example of broken encapsulation.
            # The control over such operations should belong to the class Task.
            # If it's done, from the outside of the class there will be just
            # one visible operation "continue_task()" or something like that.
            from mistral.engine import task_handler as t_h

            task.set_state(
                states.WAITING,
                "Delayed by 'retry' policy [delay=%s]" % self.delay
            )

            t_h._schedule_refresh_task_state(task.get_id(), self.delay)

            return

        task.set_state(
            states.RUNNING_DELAYED,
            "Delayed by 'retry' policy [delay=%s]" % self.delay
        )

        sched = sched_base.get_system_scheduler()

        job = sched_base.SchedulerJob(
            run_after=self.delay,
            func_name=_CONTINUE_TASK_PATH,
            func_args={'task_ex_id': task.get_id()}
        )

        sched.schedule(job)


class TimeoutPolicy(base.TaskPolicy):
    _schema = {
        "properties": {
            "delay": {
                "type": "integer",
                "minimum": 0
            }
        }
    }

    def __init__(self, timeout_sec):
        self.delay = timeout_sec

    def before_task_start(self, task):
        super(TimeoutPolicy, self).before_task_start(task)

        # No timeout if delay is 0
        if self.delay == 0:
            return

        sched = sched_base.get_system_scheduler()

        job = sched_base.SchedulerJob(
            run_after=self.delay,
            func_name=_FAIL_IF_INCOMPLETE_TASK_PATH,
            func_args={
                'task_ex_id': task.get_id(),
                'timeout': self.delay
            }
        )

        sched.schedule(job)

        wf_trace.info(
            task.task_ex,
            "Timeout check scheduled [task=%s, timeout(s)=%s]." %
            (task.get_id(), self.delay)
        )


class PauseBeforePolicy(base.TaskPolicy):
    _schema = {
        "properties": {
            "expr": {"type": "boolean"}
        }
    }

    def __init__(self, expression):
        self.expr = expression

    def before_task_start(self, task):
        super(PauseBeforePolicy, self).before_task_start(task)

        if not self.expr:
            return

        wf_trace.info(
            task.task_ex,
            "Workflow paused before task '%s' [%s -> %s]" %
            (
                task.get_name(),
                task.wf_ex.state,
                states.PAUSED
            )
        )

        task.set_state(states.IDLE, "Set by 'pause-before' policy")

        wf_handler.pause_workflow(task.wf_ex)


class ConcurrencyPolicy(base.TaskPolicy):
    _schema = {
        "properties": {
            "concurrency": {
                "type": "integer",
                "minimum": 0
            }
        }
    }

    def __init__(self, concurrency):
        self.concurrency = concurrency

    def before_task_start(self, task):
        super(ConcurrencyPolicy, self).before_task_start(task)

        if self.concurrency == 0:
            return

        # This policy doesn't do anything except validating "concurrency"
        # property value and setting a variable into task runtime context.
        # This variable is then used to define how many action executions
        # may be started in parallel.
        ctx_key = 'concurrency'

        task.set_runtime_context_value(ctx_key, self.concurrency)


class FailOnPolicy(base.TaskPolicy):
    _schema = {
        "properties": {
            "fail-on": {"type": "boolean"},
        }
    }

    def __init__(self, fail_on):
        self.fail_on = fail_on

    def before_task_start(self, task):
        pass

    def after_task_complete(self, task):
        super(FailOnPolicy, self).after_task_complete(task)

        if task.get_state() != states.SUCCESS:
            return

        if self.fail_on:
            task.set_state(states.ERROR, "Failed by 'fail-on' policy")


@db_utils.retry_on_db_error
@post_tx_queue.run
def _continue_task(task_ex_id):
    from mistral.engine import task_handler

    with db_api.transaction():
        task_ex = db_api.load_task_execution(task_ex_id)

        task_handler.continue_task(task_ex)


@db_utils.retry_on_db_error
@post_tx_queue.run
def _complete_task(task_ex_id, state, state_info):
    from mistral.engine import task_handler

    with db_api.transaction():
        task_ex = db_api.load_task_execution(task_ex_id)

        task_handler.complete_task(task_ex, state, state_info)


@db_utils.retry_on_db_error
@post_tx_queue.run
def _fail_task_if_incomplete(task_ex_id, timeout):
    from mistral.engine import task_handler

    with db_api.transaction():
        task_ex = db_api.load_task_execution(task_ex_id)

        if not states.is_completed(task_ex.state):
            msg = 'Task timed out [timeout(s)=%s].' % timeout

            task_handler.complete_task(task_ex, states.ERROR, msg)



# Copyright 2020 Nokia Software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""add namespace column to action definitions

Revision ID: 037
Revises: 036
Create Date: 2020-1-6 10:22:20

"""

# revision identifiers, used by Alembic.

from alembic import op
import sqlalchemy as sa
from sqlalchemy.engine import reflection
from sqlalchemy.sql import table, column

revision = '037'
down_revision = '036'


def upgrade():
    op.add_column(
        'action_definitions_v2',
        sa.Column('namespace', sa.String(length=255), nullable=True)
    )

    inspect = reflection.Inspector.from_engine(op.get_bind())

    unique_constraints = [
        unique_constraint['name'] for unique_constraint in
        inspect.get_unique_constraints('action_definitions_v2')
    ]
    if 'name' in unique_constraints:
        op.drop_index('name', table_name='action_definitions_v2')

    if 'action_definitions_v2_name_project_id_key' in unique_constraints:
        op.drop_constraint('action_definitions_v2_name_project_id_key',
                           table_name='action_definitions_v2')

    op.create_unique_constraint(
        None,
        'action_definitions_v2',
        ['name', 'namespace', 'project_id']
    )

    action_def = table('action_definitions_v2', column('namespace'))
    session = sa.orm.Session(bind=op.get_bind())
    with session.begin(subtransactions=True):
        session.execute(
            action_def.update().values(namespace='').where(
                action_def.c.namespace is None))  # noqa

    session.commit()



# Copyright 2020 - Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import json
import math
import random
import string

from rally.task import validation
from rally_openstack import consts
from rally_openstack import scenario
from rally_openstack.scenarios.mistral import utils


def random_string(length=10):
    """Generate a random string of given length """

    letters = string.ascii_lowercase

    return ''.join(random.choices(letters, k=length))


class MistralExpressionScenario(utils.MistralScenario):

    def run(self, tasks_number=50, params_size_mb=5):
        wf_text, wf_name = self.create_wf_string(tasks_number)
        params = self.create_params(params_size_mb)

        self._create_workflow(wf_text)
        self._create_execution(wf_name, **params)

    def create_params(self, size_mb):
        block_size_mb = 0.2
        rand_string = random_string(105400)
        number_of_fields = math.floor(size_mb / block_size_mb)
        data_list = ''
        #  each one of these blocks is 200kb
        data_template = """
                "data%s":
                {
                  "dummy": 5470438,
                  "data_value": -796997888,
                  "sub_data":
                  {
                    "meta_data":
                    {
                      "value": "%s",
                      "text": "dummy text"
                    },
                    "field1": "%s",
                    "field2": false,
                    "field3":
                    {
                      "value1": -1081872761.2081857,
                      "value2": -1081872761.2081857
                    }
                  }
                },
"""
        wf_params = """
        {
          "field": "some Value",
            "data":
              {
                {{{__DATA_LIST__}}}
              }
        }
"""
        for i in range(1, int(number_of_fields + 1)):
            data_list += data_template % (i, rand_string, rand_string)

        data_list = data_list[:-2]

        wf_params = wf_params.replace('{{{__DATA_LIST__}}}', data_list)
        params = json.loads(wf_params)

        return params

    def get_query(self):
        raise NotImplementedError

    def create_wf_string(self, tasks_number):
        wf_tasks = ''
        wf_name = 'wf_{}'.format(random_string(5))

        query = self.get_query()

        wf_text = """
            version: '2.0'
            {}:
              tasks:
                task0:
                  action: std.noop
                {{{__TASK_LIST__}}}
            """
        task_template = """
                task{}:
                  action: std.noop
                  publish:
                     output{}: {}
            """

        for i in range(1, tasks_number + 1):
            wf_tasks += task_template.format(i, i, query)

        wf_text = wf_text.replace('{{{__TASK_LIST__}}}', wf_tasks)

        wf_text = wf_text.format(wf_name)

        return wf_text, wf_name


@validation.add("required_platform", platform="openstack", users=True)
@validation.add("required_services", services=[consts.Service.MISTRAL])
@scenario.configure(name="MistralExecutions.YaqlExpression",
                    platform="openstack")
class YaqlExpressionScenario(MistralExpressionScenario):

    def get_query(self):
        return '<% data %>'


@validation.add("required_platform", platform="openstack", users=True)
@validation.add("required_services", services=[consts.Service.MISTRAL])
@scenario.configure(name="MistralExecutions.JinjaExpression",
                    platform="openstack")
class JinjaExpressionScenario(MistralExpressionScenario):

    def get_query(self):
        return '{{ data }} '



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_policy import policy

from mistral.policies import base

MEMBERS = 'members:%s'

# NOTE(hieulq): all API operations of below rules are not documented in API
# reference docs yet.
rules = [
    policy.DocumentedRuleDefault(
        name=MEMBERS % 'create',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description='Shares the resource to a new member.',
        operations=[
            {
                'path': '/v2/members',
                'method': 'POST'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=MEMBERS % 'delete',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description='Deletes a member from the member list of a resource.',
        operations=[
            {
                'path': '/v2/members',
                'method': 'DELETE'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=MEMBERS % 'get',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description='Shows resource member details.',
        operations=[
            {
                'path': '/v2/members/{member_id}',
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=MEMBERS % 'list',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description='Return all members with whom the resource has been '
                    'shared.',
        operations=[
            {
                'path': '/v2/members',
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=MEMBERS % 'update',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description='Sets the status for a resource member.',
        operations=[
            {
                'path': '/v2/members',
                'method': 'PUT'
            }
        ]
    )
]


def list_rules():
    return rules



# Copyright 2016 - Brocade Communications Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os

from cachetools import cached
from cachetools import LRUCache
import json
import jwt
from jwt import algorithms as jwt_algos
from oslo_config import cfg
from oslo_log import log as logging
import pprint
import requests
from urllib import parse

from mistral._i18n import _
from mistral import auth
from mistral import exceptions as exc


LOG = logging.getLogger(__name__)

CONF = cfg.CONF


class KeycloakAuthHandler(auth.AuthHandler):
    def authenticate(self, req):
        if 'X-Auth-Token' not in req.headers:
            msg = _("Auth token must be provided in 'X-Auth-Token' header.")

            LOG.error(msg)

            raise exc.UnauthorizedException(message=msg)

        access_token = req.headers.get('X-Auth-Token')

        try:
            decoded = jwt.decode(
                access_token,
                algorithms=['RS256'],
                verify=False
            )
        except Exception as e:
            msg = _("Token can't be decoded because of wrong format %s")\
                % str(e)

            LOG.error(msg)

            raise exc.UnauthorizedException(message=msg)

        # Get user realm from parsed token
        # Format is "iss": "http://<host>:<port>/auth/realms/<realm_name>",
        __, __, realm_name = decoded['iss'].strip().rpartition('/realms/')
        audience = decoded.get('aud')

        # Get roles from parsed token
        roles = ','.join(decoded['realm_access']['roles']) \
            if 'realm_access' in decoded else ''

        # NOTE(rakhmerov): There's a special endpoint for introspecting
        # access tokens described in OpenID Connect specification but it's
        # available in KeyCloak starting only with version 1.8.Final so we have
        # to use user info endpoint which also takes exactly one parameter
        # (access token) and replies with error if token is invalid.
        user_info_endpoint_url = CONF.keycloak_oidc.user_info_endpoint_url

        if user_info_endpoint_url.startswith(('http://', 'https://')):
            self.send_request_to_auth_server(
                url=user_info_endpoint_url,
                access_token=access_token
            )
        else:
            public_key = self.get_public_key(realm_name)

            keycloak_iss = None

            try:
                if CONF.keycloak_oidc.keycloak_iss:
                    keycloak_iss = CONF.keycloak_oidc.keycloak_iss % realm_name

                jwt.decode(
                    access_token,
                    public_key,
                    audience=audience,
                    issuer=keycloak_iss,
                    algorithms=['RS256'],
                    verify=True
                )
            except Exception:
                LOG.exception('The request access token is invalid.')

                raise exc.UnauthorizedException()

        req.headers["X-Identity-Status"] = "Confirmed"
        req.headers["X-Project-Id"] = realm_name
        req.headers["X-Roles"] = roles

    @staticmethod
    def get_system_ca_file():
        """Return path to system default CA file."""
        # Standard CA file locations for Debian/Ubuntu, RedHat/Fedora,
        # Suse, FreeBSD/OpenBSD, MacOSX, and the bundled ca.
        ca_path = [
            '/etc/ssl/certs/ca-certificates.crt',
            '/etc/pki/tls/certs/ca-bundle.crt',
            '/etc/ssl/ca-bundle.pem',
            '/etc/ssl/cert.pem',
            '/System/Library/OpenSSL/certs/cacert.pem',
            requests.certs.where()
        ]

        for ca in ca_path:
            LOG.debug("Looking for ca file %s", ca)

            if os.path.exists(ca):
                LOG.debug("Using ca file %s", ca)

                return ca

        LOG.warning("System ca file could not be found.")

    @cached(LRUCache(maxsize=32))
    def get_public_key(self, realm_name):
        keycloak_key_url = (
            CONF.keycloak_oidc.auth_url +
            CONF.keycloak_oidc.public_cert_url % realm_name
        )

        response_json = self.send_request_to_auth_server(keycloak_key_url)

        keys = response_json.get('keys')

        if not keys:
            raise exc.MistralException(
                'Unexpected response structure from the keycloak server.'
            )

        public_key = jwt_algos.RSAAlgorithm.from_jwk(
            json.dumps(keys[0])
        )

        return public_key

    def send_request_to_auth_server(self, url, access_token=None):
        certfile = CONF.keycloak_oidc.certfile
        keyfile = CONF.keycloak_oidc.keyfile
        cafile = CONF.keycloak_oidc.cafile or self.get_system_ca_file()
        insecure = CONF.keycloak_oidc.insecure

        verify = None

        if parse.urlparse(url).scheme == "https":
            verify = False if insecure else cafile

        cert = (certfile, keyfile) if certfile and keyfile else None

        headers = {}

        if access_token:
            headers["Authorization"] = "Bearer %s" % access_token

        try:
            resp = requests.get(
                url,
                headers=headers,
                verify=verify,
                cert=cert
            )
        except requests.ConnectionError:
            msg = _(
                "Can't connect to the keycloak server with address '%s'."
            ) % url

            LOG.exception(msg)

            raise exc.MistralException(message=msg)

        if resp.status_code == 401:
            LOG.warning(
                "HTTP response from OIDC provider:"
                " [%s] with WWW-Authenticate: [%s]",
                pprint.pformat(resp.text),
                resp.headers.get("WWW-Authenticate")
            )
        else:
            LOG.debug(
                "HTTP response from the OIDC provider: %s",
                pprint.pformat(resp.json())
            )

        resp.raise_for_status()

        return resp.json()



# Copyright 2013 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import datetime
import json
import pkg_resources as pkg
import time
from unittest import mock

from oslo_config import cfg
from oslo_log import log as logging
from oslotest import base

from mistral import context as auth_context
from mistral.db.sqlalchemy import base as db_sa_base
from mistral.db.sqlalchemy import sqlite_lock
from mistral.db.v2 import api as db_api
from mistral.lang import parser as spec_parser
from mistral.services import actions as action_service
from mistral.services import security
from mistral.tests.unit import config as test_config
from mistral import version

RESOURCES_PATH = 'tests/resources/'
LOG = logging.getLogger(__name__)


test_config.parse_args()


def get_resource(resource_name):
    return open(pkg.resource_filename(
        version.version_info.package,
        RESOURCES_PATH + resource_name)).read()


def get_context(default=True, admin=False):
    if default:
        return auth_context.MistralContext.from_dict({
            'user_name': 'test-user',
            'user': '1-2-3-4',
            # TODO(tkajinam): Remove this once oslo.context >= 4.0.0 becomes
            #                 avaialble
            'tenant': security.DEFAULT_PROJECT_ID,
            'project_id': security.DEFAULT_PROJECT_ID,
            'project_name': 'test-project',
            'is_admin': admin
        })
    else:
        return auth_context.MistralContext.from_dict({
            'user_name': 'test-user',
            'user': '9-0-44-5',
            # TODO(tkajinam): Remove this once oslo.context >= 4.0.0 becomes
            #                 avaialble
            'tenant': '99-88-33',
            'project_id': '99-88-33',
            'project_name': 'test-another',
            'is_admin': admin
        })


class FakeHTTPResponse(object):
    def __init__(self, text, status_code, reason=None, headers=None,
                 history=None, encoding='utf-8', url='', cookies=None,
                 elapsed=None):
        self.text = text
        self.content = text
        self.status_code = status_code
        self.reason = reason
        self.headers = headers or {}
        self.history = history
        self.encoding = encoding
        self.url = url
        self.cookies = cookies or {}
        self.elapsed = elapsed or datetime.timedelta(milliseconds=123)

    def json(self, **kwargs):
        return json.loads(self.text, **kwargs)


class BaseTest(base.BaseTestCase):
    def setUp(self):
        super(BaseTest, self).setUp()

        # By default, retain only built-in actions so that the unit tests
        # don't see unexpected actions (i.e. provided by action generators
        # installed by other projects).
        self.override_config(
            'only_builtin_actions',
            True,
            'legacy_action_provider'
        )
        self.override_config(
            'load_action_generators',
            False,
            'legacy_action_provider'
        )

        self.addCleanup(spec_parser.clear_caches)

        def _cleanup_actions():
            action_service.get_test_action_provider().cleanup()

        self.addCleanup(_cleanup_actions)

    def register_action_class(self, name, cls):
        action_service.get_test_action_provider().register_python_action(
            name,
            cls
        )

    def assertRaisesWithMessage(self, exception, msg, func, *args, **kwargs):
        try:
            func(*args, **kwargs)
            self.assertFail()
        except exception as e:
            self.assertEqual(msg, e.message)

    def assertRaisesWithMessageContaining(self, exception, msg, func, *args,
                                          **kwargs):
        try:
            func(*args, **kwargs)
            self.assertFail()
        except exception as e:
            self.assertIn(msg, e.message)

    def assertListEqual(self, l1, l2):
        super(BaseTest, self).assertListEqual(l1, l2)

    def assertDictEqual(self, cmp1, cmp2):
        super(BaseTest, self).assertDictEqual(cmp1, cmp2)

    def _assert_single_item(self, items, **props):
        return self._assert_multiple_items(items, 1, **props)[0]

    def _assert_no_item(self, items, **props):
        self._assert_multiple_items(items, 0, **props)

    def _assert_multiple_items(self, items, count, **props):
        def _matches(item, **props):
            for prop_name, prop_val in props.items():
                v = item[prop_name] if isinstance(
                    item, dict) else getattr(item, prop_name)

                if v != prop_val:
                    return False

            return True

        filtered_items = list(
            [item for item in items if _matches(item, **props)]
        )

        found = len(filtered_items)

        if found != count:
            LOG.info("[failed test ctx] items=%s, expected_props=%s", str(
                items), props)
            self.fail("Wrong number of items found [props=%s, "
                      "expected=%s, found=%s]" % (props, count, found))

        return filtered_items

    def _assert_dict_contains_subset(self, expected, actual, msg=None):
        """Checks whether actual is a superset of expected.

        Note: This is almost the exact copy of the standard method
        assertDictContainsSubset() that appeared in Python 2.7, it was
        added to use it with Python 2.6.
        """
        missing = []
        mismatched = []

        for key, value in expected.items():
            if key not in actual:
                missing.append(key)
            elif value != actual[key]:
                mismatched.append('%s, expected: %s, actual: %s' %
                                  (key, value,
                                   actual[key]))

        if not (missing or mismatched):
            return

        standardMsg = ''

        if missing:
            standardMsg = 'Missing: %s' % ','.join(m for m in missing)
        if mismatched:
            if standardMsg:
                standardMsg += '; '
            standardMsg += 'Mismatched values: %s' % ','.join(mismatched)

        self.fail(self._formatMessage(msg, standardMsg))

    def _await(self, predicate, delay=1, timeout=60, fail_message="no detail",
               fail_message_formatter=lambda x: x):
        """Awaits for predicate function to evaluate to True.

        If within a configured timeout predicate function hasn't evaluated
        to True then an exception is raised.
        :param predicate: Predicate function.
        :param delay: Delay in seconds between predicate function calls.
        :param timeout: Maximum amount of time to wait for predication
            function to evaluate to True.
        :param fail_message: explains what was expected
        :param fail_message_formatter: lambda that formats the fail_message
        :return:
        """
        end_time = time.time() + timeout

        while True:
            if predicate():
                break

            if time.time() + delay > end_time:
                raise AssertionError(
                    "Failed to wait for expected result: " +
                    fail_message_formatter(fail_message)
                )

            time.sleep(delay)

    def _sleep(self, seconds):
        time.sleep(seconds)

    def override_config(self, name, override, group=None):
        """Cleanly override CONF variables."""
        cfg.CONF.set_override(name, override, group)

        self.addCleanup(cfg.CONF.clear_override, name, group)


class DbTestCase(BaseTest):
    is_heavy_init_called = False

    @classmethod
    def __heavy_init(cls):
        """Method that runs heavy_init().

        Make this method private to prevent extending this one.
        It runs heavy_init() only once.

        Note: setUpClass() can be used, but it magically is not invoked
        from child class in another module.
        """
        if not cls.is_heavy_init_called:
            cls.heavy_init()
            cls.is_heavy_init_called = True

    @classmethod
    def heavy_init(cls):
        """Runs a long initialization.

        This method runs long initialization  once by class
        and can be extended by child classes.
        """
        # If using sqlite, change to memory. The default is file based.
        if cfg.CONF.database.connection.startswith('sqlite'):
            cfg.CONF.set_default('connection', 'sqlite://', group='database')

        cfg.CONF.set_default('max_overflow', -1, group='database')
        cfg.CONF.set_default('max_pool_size', 1000, group='database')

        db_api.setup_db()

    def _clean_db(self):
        contexts = [
            get_context(default=False),
            get_context(default=True)
        ]

        for ctx in contexts:
            auth_context.set_ctx(ctx)

            with mock.patch('mistral.services.security.get_project_id',
                            new=mock.MagicMock(return_value=ctx.project_id)):
                with db_api.transaction():
                    db_api.delete_event_triggers()
                    db_api.delete_cron_triggers()
                    db_api.delete_workflow_executions()
                    db_api.delete_task_executions()
                    db_api.delete_action_executions()
                    db_api.delete_workbooks()
                    db_api.delete_workflow_definitions()
                    db_api.delete_action_definitions()
                    db_api.delete_environments()
                    db_api.delete_resource_members()
                    db_api.delete_delayed_calls()
                    db_api.delete_scheduled_jobs()

        sqlite_lock.cleanup()

        if not cfg.CONF.database.connection.startswith('sqlite'):
            db_sa_base.get_engine().dispose()

    def setUp(self):
        super(DbTestCase, self).setUp()

        self.__heavy_init()

        self.ctx = get_context()

        auth_context.set_ctx(self.ctx)

        self.addCleanup(auth_context.set_ctx, None)
        self.addCleanup(self._clean_db)

    def is_db_session_open(self):
        return db_sa_base._get_thread_local_session() is not None



# Copyright 2015 Huawei Technologies Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log
import tenacity
import tooz.coordination

from mistral_lib import utils


LOG = log.getLogger(__name__)

_SERVICE_COORDINATOR = None


class ServiceCoordinator(object):
    """Service coordinator.

    This class uses the `tooz` library to manage group membership.

    To ensure that the other agents know this agent is still alive,
    the `heartbeat` method should be called periodically.
    """

    def __init__(self, my_id=None):
        self._coordinator = None
        self._my_id = my_id or utils.get_process_identifier()
        self._my_id = self._my_id.encode("latin-1")
        self._started = False

    def start(self):
        backend_url = cfg.CONF.coordination.backend_url

        if backend_url:
            try:
                self._coordinator = tooz.coordination.get_coordinator(
                    backend_url,
                    self._my_id
                )

                self._coordinator.start(start_heart=True)
                self._started = True

                LOG.info('Coordination backend started successfully.')
            except tooz.coordination.ToozError as e:
                self._started = False

                LOG.exception('Error connecting to coordination backend. '
                              '%s', str(e))

    def stop(self):
        if not self.is_active():
            return

        try:
            self._coordinator.stop()
        except tooz.coordination.ToozError:
            LOG.warning('Error connecting to coordination backend.')
        finally:
            self._coordinator = None
            self._started = False

    def is_active(self):
        return self._coordinator and self._started

    @tenacity.retry(stop=tenacity.stop_after_attempt(5))
    def join_group(self, group_id):
        if not self.is_active() or not group_id:
            return

        try:
            join_req = self._coordinator.join_group(group_id.encode("latin-1"))
            join_req.get()

            LOG.info(
                'Joined service group:%s, member:%s',
                group_id,
                self._my_id
            )

            return
        except tooz.coordination.MemberAlreadyExist:
            return
        except tooz.coordination.GroupNotCreated as e:
            create_grp_req = self._coordinator.create_group(
                group_id.encode("latin-1"))

            try:
                create_grp_req.get()
            except tooz.coordination.GroupAlreadyExist:
                pass

            # Re-raise exception to join group again.
            raise e

    def leave_group(self, group_id):
        if self.is_active():
            self._coordinator.leave_group(group_id.encode("latin-1"))

            LOG.info(
                'Left service group:%s, member:%s',
                group_id,
                self._my_id
            )

    def get_members(self, group_id):
        """Gets members of coordination group.

        ToozError exception must be handled when this function is invoded, we
        leave it to the invoker for the handling decision.
        """
        if not self.is_active():
            return []

        get_members_req = self._coordinator.get_members(
            group_id.encode("latin-1"))

        try:
            members = get_members_req.get()

            LOG.debug('Members of group %s: %s', group_id, members)

            return members
        except tooz.coordination.GroupNotCreated:
            LOG.warning('Group %s does not exist.', group_id)

            return []


def cleanup_service_coordinator():
    """Intends to be used by tests to recreate service coordinator."""

    global _SERVICE_COORDINATOR

    _SERVICE_COORDINATOR = None


def get_service_coordinator(my_id=None):
    global _SERVICE_COORDINATOR

    if not _SERVICE_COORDINATOR:
        _SERVICE_COORDINATOR = ServiceCoordinator(my_id=my_id)
        _SERVICE_COORDINATOR.start()

    return _SERVICE_COORDINATOR


class Service(object):
    def __init__(self, group_type):
        self.group_type = group_type

    @lockutils.synchronized('service_coordinator')
    def register_membership(self):
        """Registers group membership.

        Because this method will be invoked on each service startup almost at
        the same time, so it must be synchronized, in case all the services
        are started within same process.
        """

        service_coordinator = get_service_coordinator()

        if service_coordinator.is_active():
            service_coordinator.join_group(self.group_type)

    def stop(self):
        service_coordinator = get_service_coordinator()

        if service_coordinator.is_active():
            service_coordinator.stop()
